{ FoxClone
  Copyright (C) 2019, 2020 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT RESTORE
Contains restore specific procedures.

-------------------------------------------------------------------------------}
unit restore;

{$mode objfpc}{$H+}

interface

procedure RestBox1(Source : integer);
procedure RestBox2(Source : integer);
procedure RestBox3(Source : integer);
procedure RestGetFile(Source : integer);
procedure RestRestore;
procedure RestConfirm;

implementation
uses
  Interfaces, Classes, SysUtils, Forms, Controls, Graphics, Dialogs, ComCtrls,
  CheckLst, StdCtrls, dateutils, strutils, fileutil, process, utilities, mainform, progform, confform;
{-------------------------------------------------------------------------------
PROCEDURE RESTBOX1
Resets boxes 2 - 4 on the restore page and sets the backdrive record = the drive
containing the backup files.

Parameters:
  Source - the item clicked in the source drive checkbox

Used by: Unit MainFm - RestBox1CK

V1.0 - 15 Oct 2019
V1.1 - 14 Feb 2020 debug code
-------------------------------------------------------------------------------}
procedure RestBox1(Source : integer);
var
  i : integer;
  maxlength : integer = 0;
  partstr : string;
begin
  //debug
  if debug then DebugOut(true,'RestBox1','Source',IntToStr(Source));
  //end debug
  with MainFm do
       begin
         RestChkBox1.CheckAll(cbUnchecked);
         RestChkBox1.Checked[Source]:= true;
         RestChkBox2.Enabled:= true;
         RestChkBox2.Clear;
         //reset everything
         RestChkBox3.Clear;
         RestChkBox3.Enabled:= false;
         RestChkBox3.ScrollWidth:= ChkBoxWidth;
         RestBox3Lbl.Caption:= '3 - Contents of Backup:';
         RestLstBox4.Clear;
         RestLstBox4.Enabled:= false;
         RestLstBox4.ScrollWidth:= ChkBoxWidth;
         RestBox4Lbl.Caption:= '4 - Destination Drive:';
         NoteBtn.Enabled:= false;
         NoteBtn.Visible:= false;
         RestBtn.Enabled:= false;
         RDestInfoLbl.Visible:= false;
         RDateLbl.Visible:= false;
       end;
  BackDrive:= MyDrives[Source + 1]; //set the drive for the restore source
  //populate checkbox2 with the partitions on the source drive
  for i:= 1 to BackDrive.totalparts do
      begin
        partstr:= PartInfo(BackDrive.parts[i],BackDrive.ptabletype);
        MainFm.RestChkBox2.Items.Add(partstr);
        if length(partstr) > maxlength then maxlength:= length(partstr);
      end;
  //set scroll bars on the source partition checklistbox
  with MainFm do if maxlength * 6 > ChkBoxWidth
                 then RestChkBox2.ScrollWidth:= ChkBoxWidth + 1
                 else RestChkBox2.ScrollWidth:= ChkBoxWidth;
end; //end procedure RestBox1
{-------------------------------------------------------------------------------
PROCEDURE RESTBOX2
Checks for a valid partition type that contains the backup files. If the partition
is valid, enables the checkbox.
Runs the procedure to open the backup file and get the backup data.

Parameters:
  Source - the item clicked in the source partition checkbox

Used by: Unit MainFm - RestBox2CK

V1.0 - 15 Oct 2019
V1.1 - 14 Feb 2020 debug code
-------------------------------------------------------------------------------}
procedure RestBox2(Source : integer);
begin
  //debug
  if debug then DebugOut(true,'RestBox2','Source',IntToStr(Source));
  //end debug
  MainFm.RestChkBox2.CheckAll(cbUnchecked);
  //Cannot have backup files on an unknown or extended or swap partition
  with Backdrive do if (parts[Source+1].fs = '<unknown>')
       or (parts[Source+1].fs = 'swap')
       or (parts[Source+1].ptype = 'extended') then exit;
  MainFm.RestChkBox2.Checked[Source]:= true;
  MainFm.RestLstBox4.Clear;
  RestGetFile(Source);
end; //end procedure RestBox2
{-------------------------------------------------------------------------------
PROCEDURE RESTBOX3
Toggles the state of the item clicked and checks if the partition table has
changed since the backup. If it has changed, the partition table MUST be restored
and the entry for the partition table is 'switched' on.

RestBox3 lists the partitions in the backup selected.

V1.1 - if the partition table has not changed and it is 'unchecked', it cannot be subsequently
checked again. Code changed so that it can be checked and unchecked.
V1.2 - checks that if the block size for any partition = 0 then that item cannot be checked. This
will apply to any partition not backed up or extended partition or a swap partition.

Parameters:
Source - the item clicked in the checkbox displaying the partitions in the
         backup set

Used by: Unit mainform - RestBox3CK

V1.0 - 15 Oct 2019
V1.1 - 29 Oct 2019
V1.2 - 14 Feb 2020 debug code
-------------------------------------------------------------------------------}
procedure RestBox3(Source : integer);
begin
  //debug
  if debug then DebugOut(true,'RestBox3','Source',IntToStr(Source));
  //end debug
  with MainFm.RestChkBox3 do
       if Source = 0 then
          if not CheckPartTable(RSettings.Mountpt,RestFileName) then
             Checked[Source]:= not Checked[Source]
          else Checked[Source]:= true
       else
          if RSettings.NewParts[Source].blocks = 0 then Checked[Source]:= false
          else Checked[Source]:= not Checked[Source];
end; //end procedure RestBox3

{-------------------------------------------------------------------------------
PROCEDURE RESTGETFILE
Checks that the partition selected is mounted, if not mounts it.
Prompts for and opens the backup file.
Reads the contents of the backup file and populates the RSettings record with
the information in file.
Populates checkbox 3 with the partitions in the backup set (as defined by the
backup file).
Searches for the drive in the backup with current drives in the system. Where
the drive in the backup is found, populates box4 with the current contents of the
drive.

V1.1 - modified so that if the blocks=0 on any partition in the backup set then that item in
       the list box is not checked. For a swap partition, contents of RestChkBox3.Items changed to strip
       out the UUID.

V1.2 - adds an * to any partition that contains an operating system (as determined by os-prober)

V1.4 - checks that the destination drive is still on the same port as it was when the backup was taken,
       e.g. if the backup was on sda that the drive is still on sda and not sdx. Where the port has
       moved raises a msg dialog asking the user if they want to continue with the restore.

V1.5 - amended to check that destination for the restore matches on both the model name and the drive
       serial number. As the serial number is an empty string from a backup taken with foxclone < V40
       if the backup serial = is blank then the drive serial # is set blank for that versions later
       than V40 will work with earlier backups.
V1.6 - changes as a result of introducing TRestPartData to hold partition info instead of the previous
       single string - so that the presentation of partition info is more flexible. Also added else
       clause to the check on user selecting a backup file - if the user cancels and had made a previous
       selection, boxes 3 & 4 are not cleared nor is the selection in box 2 unchecked. fixed.
V1.7   Bug fix - if a user selects a restore from an old backup and then selects a new restore from a
       backup that stores the serial # it will generate an error - can't find drive. Because the first
       restore sets mydrives serial # = '' which then fails the next test. Fixed.
       Code changes consequent on using buildpartstr.

Parameters:
  Source - the item clicked in the source partition checkbox

Used by: RestBox2

V1.0 - 15 Oct 2019
V1.1 - 03 Nov 2019
V1.2 - 08 Nov 2019
V1.3 - 14 Feb 2020 debug code
V1.4 - 31 Mar 2020
V1.5 - 09 Jun 2020
V1.6 - 19 Sep 2020
V1.7 - 19 Sep 2020
-------------------------------------------------------------------------------}
procedure RestGetFile(Source : integer);
var
  OpenDialog : TOpenDialog;
  ForceMount : boolean = false;
  MountOK : boolean = false;
  MountPoint : string;
  i : integer;
  maxlength : integer = 0;
  Found : integer = 0;
  CheckSerial : string; //bug fix in V1.7
begin
  //debug
  if debug then DebugOut(true,'RestGetFile','Source',IntToStr(Source));
  //end debug
  MountPoint:= CheckMount(BackDrive.parts[Source+1].pnumber);
  if MountPoint <> '<none>' then MountOK:= true
  else
    begin
      if mkdir(BackDrive.parts[Source+1].pnumber) = 0 then
         begin
           if mount(BackDrive.parts[Source+1].pnumber,true) = 0 then
              begin
                MountPoint:= MountFolder + BackDrive.parts[Source+1].pnumber;
                MountOK:= true;
                ForceMount:= true;
              end
         end
    end;
  if not MountOK then exit;
  //everything okay so get the backup file
  OpenDialog:= TOpenDialog.Create(MainFm);
  OpenDialog.Title:= 'Open backup file';
  OpenDialog.InitialDir:= mountpoint;
  OpenDialog.Filter := 'Backup files (*.backup)|*.backup';
  if OpenDialog.Execute then
     begin
       RestFileName:= OpenDialog.Filename;
       getRestoreData(RestFileName,RSettings);
       with MainFM do
          begin
            RestBox3Lbl.Caption:= '3 - Backup: ' + chopstr(RestFileName,45);
            RestChkBox3.Enabled:= true;
            RestChkBox3.Clear; //in case it has already been populated
            RDateLbl.Visible:= true;
            RDateLbl.Caption:= 'Backup Date:  ' + RSettings.BackDate + LF
                               + 'Time Now:       ' + FormatDateTime(FDateTimeStr,Now);
            RestChkBox3.Items.Add('Restore partition table');
            if RSettings.PTable then RestChkBox3.Checked[0]:= true;
            for i:= 1 to RSettings.TotalParts do
                begin
                  RestChkBox3.Items.Add(BuildPartStr(RSettings.NewParts[i]));
                  if RSettings.NewParts[i].blocks = 0 then RestChkBox3.Checked[i]:= false
                  else RestChkBox3.Checked[i]:= true;
                  if length(RestChkBox3.Items[i-1]) > maxlength then
                     maxlength:= length(RestChkBox3.Items[i-1]);
                end;
            //set horizontal scroll bar if needed
            if maxlength * 6 > RestChkBox3.Width then RestChkBox3.ScrollWidth:= RestChkBox3.Width + 1
            else RestChkBox3.ScrollWidth:= RestChkBox3.Width;
            //if notes in backup make note button visible
            if RSettings.Note then
               begin
                 NoteBtn.Enabled:= true;
                 NoteBtn.Visible:= true;
               end;
            //amended for foxclone V40 to include the serial # in the test
            for i:= 1 to TotalDrives do
                begin
                  //if no serial # in backup then set drive serial # blank
                  if RSettings.Serial = '' then CheckSerial:= '' else CheckSerial:= MyDrives[i].serial;
                  if ((MyDrives[i].Model = RSettings.Model) and (CheckSerial = RSettings.Serial))
                  then Found:= i;
                end;
            if Found <> 0 then
               begin
                 RestBox4Lbl.Caption:= '4 - Destination Drive: ' + RSettings.Model + ' ('
                                       + MyDrives[Found].size + ')';
                 maxlength:= 0;
                 for i:= 1 to MyDrives[Found].totalparts do
                     begin
                       RestLstBox4.Items.Add(PartInfo(MyDrives[Found].parts[i],
                                             MyDrives[Found].parts[i].ptype));
                       if length(RestLstBox4.Items[i-1]) > maxlength
                            then maxlength:= length(RestLstBox4.Items[i-1]);
                     end;
                 if maxlength * 6 > ChkBoxWidth then RestLstBox4.ScrollWidth:= ChkBoxWidth + 1
                 else RestLstBox4.ScrollWidth:= ChkBoxWidth;
                 //need to check that the restore drive is still on the same port as it was when it was
                 //backed up, e.g. if the backup was of sda, then that drive is still on sda.
                 RSettings.DestMountpt:= MyDrives[Found].Mountpt;
                 RestLstBox4.Enabled:= true;
                 RDestInfoLbl.Visible:= true;
                 if RSettings.DestMountpt = RSettings.Mountpt then RestBtn.Enabled:= true else
                    if MessageDlg('Drive ' + RSettings.Model + ' was on ' + RSettings.Mountpt
                                 + ' when the backup was taken. It is now on '
                                 + RSettings.DestMountpt + '.' + LF + 'Continue with restore to '
                                 + RSettings.DestMountpt + '?',mtWarning,[mbNo,mbYes],0) = mrNo
                    then RestBtn.Enabled:= false else RestBtn.Enabled:= true;
               end
            else MessageDlg('Cannot find drive: ' + RSettings.Model,mtError,[mbOK],0);
          end
     end
  else //user has cancelled, have to clear boxes 3 & 4 and uncheck the partition selected in box 2
     with MainFm do
          begin
            RestChkBox2.CheckAll(cbUnchecked);
            RestChkBox3.Clear;
            RestChkBox3.Enabled:= false;
            RestChkBox3.ScrollWidth:= ChkBoxWidth;
            RestBox3Lbl.Caption:= '3 - Contents of Backup:';
            RestLstBox4.Clear;
            RestLstBox4.Enabled:= false;
            RestLstBox4.ScrollWidth:= ChkBoxWidth;
            RestBox4Lbl.Caption:= '4 - Destination Drive:';
          end;
  if ForceMount then
     begin
       //unmount the partition
     end;
  OpenDialog.Free;
end; //end procedure RestGetFile
{-------------------------------------------------------------------------------
PROCEDURE RESTCONFIRM

Adjusts the confform, populates it with restore data and seeks confirmation
to proceed with the restore.

V1.1 - DestFileLbl chopped to split the line into two.

Used by: Unit mainform - RestRestoreCK (the restore button on the restore tab)

V1.0 - 15 Oct 2019
V1.1 - 04 Nov 2019
V1.2 - 14 Feb 2020 debug code
-------------------------------------------------------------------------------}
procedure RestConfirm;
var
  i : integer;
  maxlength : integer = 0;
begin
  //debug

  if debug then DebugOut(true,'RestConfirm');
  //end debug
  with ConfirmForm do
     begin
       ConfirmForm:= Tconfform.create(MainFm);
       //populate the confirm form
       Caption:= 'Confirm Restore';
       DestLbl.Visible:= false; //not needed
       DestFileLbl.Caption:= 'From: ' + ChopStr(RestFileName,40);
       SourceLbl.Caption:= 'To drive: ' + RSettings.Model;
       PartListBox.Clear;
       if MainFm.RestChkBox3.Checked[0] then PartListBox.Items.Add(MainFm.RestChkBox3.Items[0]);
       for i:= 1 to MainFm.RestChkBox3.Count - 1 do
           if MainFm.RestChkBox3.Checked[i]
              then PartListBox.Items.Add('Restore ' + MainFm.RestChkBox3.Items[i]);
       for i:= 0 to PartListBox.Items.Count - 1 do
           if length(PartListBox.Items[i]) > maxlength
              then maxlength:= length(PartListBox.Items[i]);
       if maxlength * 6 > PartListBox.Width
          then PartListBox.ScrollWidth:= PartListBox.Width + 1;
       //re-adjust the view of the form
       cDriveImg.top:= 105;
       cFileImg.top:= 1;
       cArrowImg.top:= 48;
       DestFileLbl.top:= 20;
       SourceLbl.top:= 65;
       PartListBox.top:= 85;
       OkBtn.top:= 175;
       cCancelBtn.top:= 175;
       //set action to restore and show
       CurrentAction:= acRest;
       Show;
     end;
end; //end procedure RestConfirm
{-------------------------------------------------------------------------------
PROCEDURE RESTRESTORE
Does the restore.

To do: Rsettings.ptable is redundant (and is set TRUE) it is simply recording whether the ptable is
selected in the settings tab, the ptable is always backed up and always restored if it has changed
(along with grub).

Used by: Unit confform - OkBtnCK

V1.0 - 15 Oct 19
V1.1 - 28 Oct 19 : restore swap partition added
V1.2 - 3 Nov 19 : completion code added so appearance of the progress form is the same for backup,
restore and clone when the activity is complete. Code modified to take account of ALL partitions being
listed, not just the ones being restored. Missing parts or parts such as extended and swap and now
clearly shown in the progress window.
V1.3 - 25 Nov : restoring the part table and grub even when the part table has not been selected.
V1.4 - 14 Feb 2020 : debug code
V1.5 - 24 Sep 2020 : changes consequent on introducing the TRestPartData record and replacing the parts
                     str array with newparts.
V1.6 - 15 Jan 2021 : testing starttime:= time was returning 31 dec 99, whereas now returned the correct
                     datetime, changed. incrementing blocks moved out of if clause. if clause modified so
                     that if 'cloned successfully' not detected it says finished not complete.
V1.7 - 21 Jan 2021 : added the view log function
-------------------------------------------------------------------------------}
procedure RestRestore;
const
  separator = '-----------------------------------------------------------------------------';
var
  cmdstr : string; //restore command line
  StartTime, PstartTime : TDateTime;
  TotalBlocks : int64 = 0; //total blocks to restore
  PreviousBlocks : int64 = 0; //blocks already restored
  i, j : integer;
  PartStr : string;
  UmountOK : boolean;
  complete : string;
  LogText, LogText2 : TStringList;
  Destination : string;
  LogFname : string;
  LogFiles : TStringList;
begin
  //debug
  if debug then DebugOut(true,'RestRestore');
  //end debug
  StartTime:= Now;
  ConfirmForm.Close;
  //disable the main form
  MainFm.PageCtrl.Enabled:= false;
  ProgressFm:=TProgressFm.Create(MainFm);
  //Hide the add note button, not needed for restore
  ProgressFm.AddBtn.Enabled:= false;
  ProgressFm.AddBtn.Visible:= false;
  //delete log files in the default folder, doesn't look like deletefile accepts wildcards
  LogText:= TStringList.create;
  FindAllFiles(LogText, DefaultFolder, '*', false);
  for i:= 0 to LogText.count - 1 do DeleteFile(LogText[i]);
  LogText.free;
  //Enable the view log button
  ProgressFm.LogBtn.Enabled:= false;
  ProgressFm.LogBtn.Visible:= true;
  //Dangerous to cancel a restore partway through so disable ability to cancel
  ProgressFm.CancelBtn.Enabled:= false;;
  ProgressFm.CancelBtn.Visible:= false;;
  ProgressFm.Show;
  //total blocks to restore
  for i:= 1 to RSettings.TotalParts do
      if MainFm.RestChkBox3.Checked[i] then TotalBlocks+= RSettings.NewParts[i].blocks;
  ProgressFm.OverallProgBar.Max:= TotalBlocks;
  //check if the partition table has changed since the backup
  if RSettings.PTable then ProgressFm.PartListBox.Items.Add('Partition table - not started')
  else ProgressFm.PartListBox.Items.Add('Partition table - not selected');
  //populate the list box with # parts
  for i:= 1 to RSettings.TotalParts do
      ProgressFm.PartListBox.Items.Add('Part ' + IntToStr(i) + ' - not started');
  if MainFm.RestChkBox3.Checked[0] then ProgressFm.Caption:= 'Restoring partition table'
  else ProgressFm.PartListBox.Items[0]:= 'Partition table - not selected';
  application.processmessages;
  if MainFm.RestChkBox3.Checked[0] then //the partition table has changed
     begin
       //restore the partition table and grub
       ProgressFm.PartListBox.Items[0]:= 'Partition table - started';
       application.processmessages;
       //restore grub, use the destination mount point - previously was mountpt in backup
       GrubRestore(ReplaceStr(RestFileName,'.backup','.grub'),RSettings.DestMountpt,true);
       //this gets back the part table and grub on MBR drives but with GPT part tables
       //it only restores the primary partition table at the start of the drive, the
       //backup part table at the end of the drive is unchanged and use of utilities
       //such as gdisk will throw an error. Use sfdisk to restore both primary and
       //backup part tables - use the destination mount point - previously was mountpt in backup
       if not PTableRestore(ReplaceStr(RestFileName,'.backup','.sfdisk'),RSettings.DestMountpt)
       then exit;
       ProgressFm.PartListBox.Items[0]:= 'Partition table - complete ';
       //copy the part table from the source into the default folder for the log files
       CopyFile(ReplaceStr(RestFileName,'.backup','.sfdisk'),DefaultFolder + 'clone.sfdisk');
       application.processmessages;
     end;
  //now run partclone
  for i:= 1 to MainFm.RestChkBox3.Count - 1 do if MainFm.RestChkBox3.Checked[i] then
     begin
       PstartTime:= Now;
       //set current progress bar
       ProgressFm.Caption:= 'Restoring Part ' + IntToStr(i)
                            + ' of ' + IntToStr(RSettings.TotalParts);
       //set the caption in the listbox
       ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + ' - started';
       //build the partclone command line
       writeln('Restore parameters');
       writeln('Rest filename=',RestFilename);
       writeln('Partition=',RSettings.NewParts[i].pnumber);
       writeln('Filesystem=',RSettings.NewParts[i].fs);
       cmdstr:= PartCloneR(RestFilename,RSettings.NewParts[i].pnumber,RSettings.NewParts[i].fs);
       writeln('restore cmd line=',cmdstr);
       //need to umount the partition if mounted
       //PartStr:= copy(RSettings.Parts[i],1,pos(' : ',RSettings.Parts[i])-1);
       PartStr:= RSettings.NewParts[i].pnumber;
       UmountOK:= CheckMount(PartStr) = '<none>';
       if not UmountOK then if umount(PartStr) = 0 then UmountOK:= true;
       if UmountOK then //partition umounted, safe to proceed
          begin
            //run partclone and update progress form
            RunPartClone(cmdstr,RSettings.NewParts[i].blocks,PreviousBlocks,TotalBlocks, StartTime);
            PreviousBlocks+= RSettings.NewParts[i].blocks;
            if ProgressFm.CurrentProgLbl.Caption = 'Cloned successfully' then complete:= ' - complete '
            else complete:= ' - finished ';
            ProgressFm.PartListBox.Items[i]:= 'Part '+ IntToStr(i) + complete
                                              + FormatDateTime('hh:nn:ss',Now - PstartTime);
            Logtext:= TStringList.create;
            with RSettings do Destination:= ReplaceStr(NewParts[i].pnumber,Mountpt,DestMountpt);
            LogFname:= DefaultFolder + FormatDateTime('yyyymmdd',Now)
                            + '.' + Destination + '.log';
            //writeln('restore logfile name=',LogFname);
            Logtext.loadfromfile(LogFname);
            Logtext.insert(0,(copy(LogFname,rpos('/',LogFname)+1,99)));
            Logtext.insert(0,(FormatDateTime('dd mmm yyyy, hh:nn',Now)));
            Logtext.Add(separator);
            Logtext.savetofile(LogFname);
            Logtext.free;
           end
       else //not safe to restore partition, unable to umount it
          begin
            ProgressFm.PartListBox.Items[i]:= 'Part '+IntToStr(i)+ ' - not restored';
            MessageDlg('Failed un-mounting ' + BackDrive.parts[i].pnumber + LF +
                       'Not safe to restore a live partition, skipping...',mtError,[mbOK],0)
          end;
     end
  else
    begin
      //if pos('extended',RSettings.Parts[i]) > 0 then
      if RSettings.NewParts[i].ptype = 'extended' then
         ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + ' - extended (skipped)'
      else
         if RSettings.NewParts[i].fs = 'swap' then
        //if pos('swap',RSettings.Parts[i]) > 0 then //recreate swap partition
           begin
             if RSettings.SwapPart <> '' then RestoreSwap(RSettings.SwapPart,RSettings.SwapUUID);
                ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + ' - swap recreated';
           end
        else ProgressFm.PartListBox.Items[i]:= 'Part ' + IntToStr(i) + ' - not included';
    end;
  //collect the logfiles together for viewing via the view log button
  LogText:= TStringList.create;
  //load the partition table
  if FileExists(DefaultFolder + 'clone.sfdisk') then LogText.loadfromfile(DefaultFolder + 'clone.sfdisk')
  else LogText.add('Partition table not restored');
  LogText.insert(0,'PARTITION TABLE');
  LogText.insert(0,separator);
  LogText.insert(0,'Restore ' + FormatDateTime('dd mmm yyyy, hh:nn',Now));
  LogText.add(separator);
  LogFiles:= TStringList.create;
  FindAllFiles(LogFiles, DefaultFolder, '*.log', false);
  LogFiles.sort;
  //load all the log files into a verify file
  if LogFiles.count >= 1 then
     for i:= 0 to LogFiles.count - 1 do
         begin
           LogText2:= TStringList.create;
           LogText2.loadfromfile(LogFiles[i]);
           for j:= 0 to LogText2.count - 1 do LogText.add(LogText2[j]);
           LogText2.free;
         end;
  LogText.savetofile(DefaultFolder + FormatDateTime('yyyymmdd',Now) + '.clone.log');
  LogText.free;
  LogFiles.free;
  //enable the log button
  ProgressFm.LogBtn.Enabled:= true;
  //reset the progress form to indicate completion
  ProgressFm.Caption:= 'Restore complete';
  ProgressFm.PartcloneLbl.Caption:= '';
  ProgressFm.CancelBtn.Enabled:= false;
  ProgressFm.FinishBtn.Enabled:= true;
end; //end procedure RestRestore

end.

