{ FoxClone
  Copyright (C) 2019 - 2021 Andy Hardwick
  This source is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.
  This code is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
  details.
  A copy of the GNU General Public License is available on the World Wide Web
  at <http://www.gnu.org/copyleft/gpl.html>. You can also obtain it by writing
  to the Free Software Foundation, Inc., 51 Franklin Street - Fifth Floor,
  Boston, MA 02110-1335, USA.
}

{-------------------------------------------------------------------------------
UNIT - UTILITIES

Various procedures/functions providing a range of utility functions. A significant
number are running various terminal commands to get disk/partition information.

Also contains the record array variables that hold drive information.

need to check if mountpt is still used and remove
need to check that show and selected and initialised

Note - right margin set at 105 for print from xed with a 25mm left and 5mm right margin

Note - lazarus will throw a lot of hints on compilation that 'xxProcess' does not seem to be initialised
This is because a function/procedure calling TermCmd has to pass a Tprocess as part of the command.
TermCmd then creates the process - hence the hint, lazarus can't find the creation in the
parent function/procedure. It remains the responsibility of the calling function/procedure to free the
process once finished with it. Saves a few lines of code.

V1.0 - 13 Oct 19
v1.1 - 21 Oct 19, TRestData record amended to include swap partition and swap UUID
V1.2 - 28 Oct 19, RestoreSwap procedure added
V1.3 - 04 Nov 19, function ChopStr added
V1.4 - 06 Nov 19, VersionDate constant added and function GetVersion added.
V1.5 - 07 Nov 19, OS added into Tpartparams and procedure OSprobe added.
V1.6 - 23 Nov 19, icons moved to /usr/share/pixmaps/foxclone/
V1.7 - 04 Jan 20, debug switch = boolean added to vars
V1.8 - 07 Jan 20, parted modified to trap any drives where part table is not GPT or MSDOS
                  GrubBackup added and GrubRestore modified
v1.9 - 12 Jan 20, getdriveinfo deleted, this now done by readdrives in mainform
v1.10 - 30 Jan 20, const for fox icon image added
v1.11 - 06 Feb 20, bug fixing on grub backup and restore where a clone tries to use dd to copy the
                   whole drive and not just the space to the first partition. function grubSize
                   introduced.
v1.12 - 08 Feb 20, sizeinkb modified, was not checking for kB
V1.13 - 23 Feb 20, function checkowner added. Users were trying to save backups onto a partition
                   owned by root, which of course wasn't working, checkowner added to test for this.
V1.14 - 28 Feb 20, Mountfolder const added and used for all mounts, Defaultfolder now used for
                   dumping files.
V1.15 - 18 Mar 20, default mount folder changed from /media/foxclone to /mnt/foxclone. When installed
                   from deb in LM19.3 cinnamon the default settings for nemo (file manager) meant it was
                   auto opening devices in /media and throwing up errors as foxclone was mounting and
                   unmounting partitions on start to read sizes. Foxclone still running ok so
                   this fix is to prevent problems in the parent OS.
V1.16 - 20 Mar 20, BootDrive var added. Is set by readdrives on start, contains the boot drive, e.g. sda
                   and is used to check that user not trying to back up the boot drive or clone to it.
V1.17 - 31 Mar 20, DestMountpt added to TRestData to cater for a drive moving its mount point between
                   doing a backup and then doing a restore.
V1.18 - 14 Apr 20, parted procedure modified to check if the flag on a partition is msftres and if so
                   then set the partition type to <unknown>. This is the MS reserved part on a win10
                   install and should be backed up with dd. On an pcie nvme drive it was showing as ext4.
V1.19 - 27 Apr 20, oldsettings var added so that after backup to a fat32 partition, the settings can be
                   put back to what they were before the backup. Backup to a fat32 partition auto sets
                   split files on and file size to 2048MB.
V1.20 - 6 June 20, getSerials, sataserial and nvmeserial added to get the serial # of the drives in the
                   system. TDrive record amended to add serial #.
V1.21 - 5 Sept 20, warning msg on sataserial removed - confusing some users. parted reworked to include
                   loop partition devices (a drive with no part table but a valid partition). Clicking on
                   on a drive with a loop part table is now checked in backbox1. chopstr changed to chop
                   the string at "/" starting from the left.
V1.22 - 24 Sep 20, Added TRestPartData record replacing the previous partstr array and deleting the
                   blocks array in TRestData. New procedures/functions:
                     procedure ParsePartData
                     function BuildPartStr
                     function WritePartInfo
                   to handle data in TRestPartData
                   pkexec added to cmd line in getlabel - without wasn't reading labels with blkid
                   correctly after a restore.
V1.22 - 03 Oct 20, swapon procedure added.
V1.23 - 14 Jan 21, partcloneV procedure created
-------------------------------------------------------------------------------}
unit utilities;

{$mode objfpc}{$H+}

interface

uses
  cthreads, Classes, Process, sysutils, dateutils, strutils,
  dialogs, controls, checklst, clocale;

const
  Version = 'Version 0.46';
  VersionDate = '22 Aug 2021';
  LF = LineEnding;
  DefaultFolder = '/tmp/foxclone/';
  MountFolder = '/mnt/foxclone/';
  ImageFolder = '/usr/share/pixmaps/foxclone/';
  DriveImgFile = ImageFolder + 'drive80.png';
  PartImgFile = ImageFolder + 'partition64.png';
  FileImgFile = ImageFolder + 'file2-64.png';
  ArrowImgFile = ImageFolder + 'arrow.png';
  DnArrowImgFile = ImageFolder + 'downarrow.png';
  SettingsImgFile = ImageFolder + 'settings64.png';
  FoxImgFile =  ImageFolder + 'fox-icon96.png';
  CheckImgFile = ImageFolder + 'checkbox.png';

type
  TActionType = (acBack, acRest, acClone, acFileClone, acVerify);
  TSetup = record
    parttable : boolean; //save partition table on backup
    compress  : boolean; //compress backup files
    complevel : integer; //compression level 1 - 9 where 0
    split     : boolean; //split backup files
    splitsize : integer; //max size of split files in MB
    fat32     : boolean; //set true if backup destination is fat32
  end;
  TRestPartData = record //introduced in V42, to hold the text string in a .backup file that describes a
                         //partition
    pnumber : string;  // sda1
    psize   : string;  // 30.8GB
    pname   : string;  // partition name on a GPT drive
    plabel  : string;  // [mylabel]
    fs      : string;  // ext4
    pused   : string;  // 43% used
    flags   : string;  // Flags-boot
    ptype   : string;  // only MSDOS, primary, extend, logical
    OS      : string;  // *Linux Mint 20 Ulyana*
    blocks  : int64;   // used 1K blocks in the partitions
  end;

  TRestData = record             //settings recorded in the *.backup file
    BackDate : string;           //date of backup
    PTable : boolean;            //was partition table explicitly backed up
    Compress : boolean;          //was file compression on
    Split : boolean;             //was split files on
    Note : boolean;              //is there a note file in the backup
    Model : string;              //model name of backed up drive
    Serial : string;             //serial number of the backed up drive added in foxclone v40
    SwapPart : string;           //eg sda5
    SwapUUID : string;           // UUID of the swap partition
    Mountpt : string;            //e.g. sda
    DestMountpt : string;        //e.g. sdc, the mount point of the destination drive on restore
    TotalParts : integer;        //could be removed and replaced with length(MyDrives[i].parts)
    //Parts : array of string;     //string of data for display in checkboxes
    NewParts : array of TRestPartData;
    //Blocks : array of int64;     //used 1K blocks in the partitions
  end;
  TPartparams = record
    pnumber  : string;   //e.g. sda1
    psize    : Int64;    //size of partition in k
    pused    : Int64;    //used blocks in partition in k
    ptype    : string;   //msdos only, e.g. primary, extended, logical
    fs       : string;   //e.g. ntfs, ext4, or not
    pname    : string;   //gpt only
    plabel   : string;   //if it has one
    OS       : string;   //operating system populated by OSprobe
    flags    : string;   //e.g. boot, msftdata
    selected : boolean;  //selected partition to backup to
  end;

  TDrive = record
    show       : boolean; //whether to display the drive
    model      : string;  //e.g. ATA WDC WDS240G2G0A
    serial     : string;  //the serial number of the drive
    mountpt    : string;  //e.g. sda
    size       : string;  //size of drive
    ptabletype : string;  // msdos or gpt
    totalparts : integer; //total number of partitions on drive
    parts      : array of Tpartparams;
  end;

  Tdrives = array of Tdrive; //dynamic arrays holding drive & part info

{-------------------------------------------------------------------------------
 GLOBAL VARIABLES
-------------------------------------------------------------------------------}
var
 settings, oldsettings  : TSetup; //oldsettings is used if a fat32 partition is used as a destination
                                  //for a backup, if fat32 settings are automatically set to split files.
                                  //oldsettings is used to set the settings back to what they were.
 MyDrives  : Tdrives;                                   //record of all drives
 BackDrive, DestDrive, RestDrive, CloneDrive : Tdrive;  //the drive to be backed, destination drive
                                                       //and restore drive
 TotalDrives : integer = 0;     //could be removed and replaced with length(MyDrives)
 BackFileName : string;
 RestFileName, VerifyFileName : string;
 RSettings, VSettings : TRestData;
 CurrentAction : TActionType;
 FDateTimeStr : string;                                //the format for date/time based on current locale
 debug : boolean = false; //switch for sending debug output to terminal
 BootDrive : string = ''; //contains the drive, e.g. sda, that foxclone booted from, set by ReadDrives


{-------------------------------------------------------------------------------
 PROCEDURES & FUNCTIONS
-------------------------------------------------------------------------------}
function SizeInKB(sizestr : string) : int64;  //returns no of 1k blocks from a string, e.g. 3.4MB
procedure termCmd(CmdLine: string; var Process : TProcess); //runs a terminal command with bash
function CheckMount(partition : string) : string;
function rmdir(partition : string) : integer;  //removes a folder
function mkdir(partition : string) : integer;  //creates a folder to mount a partition in
procedure parted(var drives : Tdrives; var dcount : integer); //get drive/partition info
function mount(partition : string; ReadWrite : boolean) : integer; //mounts a partition
procedure dfSizeUsed(partition : string; var size :int64; var used : int64);
function getLabel(partition :string) : string; //returns label, if one, for specified partition
procedure partLabels(var drives : Tdrives; var maxdrives : integer);
function umount(partition : string) : integer; //un mounts a partition
procedure GetPartSize(var drives : Tdrives; var maxdrives : integer);
function PartInfo(var partition : Tpartparams; parttype : string) : string;
function PercentUsed(a,b : Int64) : string;
function BlocksToStr(ablocks : int64) : string;
function ValidPartType(var parts : Tpartparams) : boolean; //checks if a partition is ok to backup to
procedure InitialSettings;
function GetPartcloneInfo(var PartCloneStr : string) : string;
function PartCloneB(partition : string; filesys : string) : string;
function PartCloneR(RestFname : string; PartNum : string; fs : string) : string;
procedure getRestoreData(FileName : string; var Settings : TRestData);
procedure ShowDrives(var checklistbox : TCheckListBox);
function CheckPartTable(device : string; fname : string) : boolean;
Procedure RunPartClone(cmdstr: string; PartUsedBlks : int64; PreviousBlocks : int64;
                       TotalBlocks : int64; StartTime : TDateTime);
function DriveSize(Drive : string) : int64;
function PartCloneC(Source : string; Dest : string; fs : string) : string;
procedure GrubRestore(Source : string; Target : String; FileFlag : boolean);
function PTableRestore(Source : string; Target : string) : boolean;
function ModifyPTable(InFile : string; OutFile : string; Target : string) : boolean;
function getUUID(partition : string) : string;
procedure RestoreSwap(partition : string; UUID : string);
function ChopStr(AString : string; Chop : integer) : string;
function GetVersion : string;
procedure OSprobe;
procedure DebugOut(proc : boolean; fname : string;
                   p1 : string = ''; p1text : string = '';
                   p2 : string = ''; p2text : string = '';
                   p3 : string = ''; p3text : string = '';
                   p4 : string = ''; p4text : string = '';
                   p5 : string = ''; p5text : string = '';
                   p6 : string = ''; p6text : string = ''); //dumps output to terminal
procedure GrubBackup(MountPoint : string; Filename : string);
function grubSize(Drive : string) : string;
function FreeSpace(var parts : Tpartparams) : boolean;
procedure getSerials(var drives : Tdrives; var dcount : integer); //get drive serial numbers
function SataSerial(MountPoint : string) : string; //returns the serial number of a sata drive
function NVMESerial(MountPoint : string) : string; //returns the serial number of an NVME drive
function IsBoot(MountPoint : string) : boolean;
procedure ParsePartData(PartStr : String; var PartData : TRestPartData);
function BuildPartStr(PartData : TRestPartData) : string;
function WritePartInfo(var partition : Tpartparams; parttype : string) : string;
procedure SwapOn(OnOff : boolean);
function PartCloneV(RestFname : string; PartNum : string) : string;

implementation

uses
  progform, forms, mainform;
{-------------------------------------------------------------------------------
FUNCTION SIZEINKB
Takes a string in the format, e.g. 34.8GB and converts it into an integer in KB
Assumes the string is in MB, GB or TB.

Amended to include kB.

Note - it is called on all drives/partitions and is used to populate the size and size
used for each partition. For partitions with known filesystems, the values generated by sizeinkb
are subsequently overwritten by output from df.

Parameters:
  sizestr - a string in the format 12.34GB or 500MB

Used by: parted - to convert reported partition sizes into 1K blocks

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code added
V1.2 - 08 Feb 20 - on a GPT drive with a bios_grub partition the size is small and measured in kB
                   not picked up. Code amended.
-------------------------------------------------------------------------------}
function SizeInKB(sizestr : string) : int64;
const
  MULTIPLY : integer = 1000; // 1000 or 1024 - need to sort this
var
  SizeUnit : string;
  Tempnum : real;
begin
  //debug
  if debug then DebugOut(false,'SizeInKB','sizestr',sizestr);
  //end debug
  if sizestr ='' then
     begin
       SizeInKB:=0;
       exit;
     end;
  SizeUnit:= copy(sizestr,length(sizestr)-1,2);
  Tempnum:= StrToFloat(copy(sizestr,1,length(sizestr)-2));
  if SizeUnit = 'kB' then SizeInKB:= round(Tempnum);
  if SizeUnit = 'MB' then SizeInKB:= round(Tempnum * MULTIPLY);
  if SizeUnit = 'GB' then SizeInKB:= round(Tempnum * MULTIPLY * MULTIPLY);
  if SizeUnit = 'TB' then SizeInKB:= round(Tempnum * MULTIPLY * MULTIPLY * MULTIPLY);
end; //end function sizeinkb
{-------------------------------------------------------------------------------
PROCEDURE TERMCMD
Executes a terminal command. To allow for piping all commands are created and
run from bash. The procedure creates the process passed but it is the
responsibility of the calling process to parse either the command's stdout or
stderr and free it on completion.

Parameters:
  CmdLine - a string containing the command line to be executed
  Process - a TProcess to be used by calling process to read output either from
            stdout or stderr

Used by:
  Unit utilities - CheckMount, rmdir, mkdir, parted, mount, dfSizeUsed,
  getLabel, umount, CheckPartTable, RunPartClone, DriveSize, GrubRestore, PTableRestore, getUUID
  RestoreSwap, GetVersion, OSprobe, GrubBackup, grubSize
  Unit backup - BackBackup
  Unit clone - CloneCheck, CloneDiskToDisk

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
procedure termCmd(CmdLine: string; var Process : TProcess);
const
  execCmd = '/bin/bash';
begin
  //debug
  if debug then DebugOut(true,'termCmd','CmdLine',CmdLine);
  //end debug
  Process:= TProcess.Create(nil);
  with Process do
       begin
         Executable:= execCmd;
         Parameters.Add('-c');
         Parameters.Add(CmdLine);
         Options:= [poUsePipes]; //to get output
         Execute; //run
       end;
 end; //end procedure TermCmd

{-------------------------------------------------------------------------------
FUNCTION CHECKMOUNT
Checks if a partition is mounted. If the partition is already mounted the function
returns the mount point as a string. If partition is not mounted the function
returns <none>.

Parameters:
  partition - a string defining the partition in the form sda2 or sdb4, etc.

Used by:
  Unit utilities - GetPartSize, PTableRestore, CheckRoot
  Unit backup - BackBackup
  Unit restore - RestGetFile, RestRestore
  Unit clone - CloneDiskToDisk, CloneFileToDisk
  Unit mainform - BackSaveCK, QuitCK

v1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
function CheckMount(partition : string) : string;
const
  bufsize = 256; // Buffer size for reading output
var
  chkProcess : TProcess;
  Buffer     : string;
  BytesRead  : integer;
  Outputstr : string = '';
begin
  //debug
  if debug then DebugOut(false,'CheckMount','partition',partition);
  //end debug
  setlength(buffer,bufsize);
  termCmd('mount | grep ' + partition,chkProcess);
  repeat
    BytesRead:= chkProcess.Output.Read(buffer[1],length(buffer));
    setlength(buffer,BytesRead);
    OutputStr+= buffer;
  until BytesRead = 0;
  if length(OutputStr) = 0 then Checkmount:= '<none>' else
     begin
       ///output is of form: /dev/sdc2 on /home type ext4 (rw,relatime,data=ordered)
       CheckMount:= copy(OutputStr,pos(' on ',OutputStr)+4,length(OutputStr));
       CheckMount:= copy(CheckMount,1,pos(' type',CheckMount)-1);
     end;
  chkProcess.free;
end; //end function checkmount
{-------------------------------------------------------------------------------
FUNCTION RMDIR
Removes a folder in the default location with the name passed as partition.
The function returns:
  0 = success.
 >0 = failure, which will normally be because the folder doesn't exist or that
      the folder is not empty.
rmdir is used to delete a folder that has been created to mount a partition.

Parameters:
  partition - a string containing the partition in the form sda1, sdc3, etc.

Used by:
  Unit utilities - GetPartSize, CheckRoot
  Unit mainform - BackSaveCK

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
V1.2 - 28 Feb 2020 - changed to pkexec rmdir as mount folder is now in /media and owned by root
-------------------------------------------------------------------------------}
function rmdir(partition : string) : integer;
const
  bufsize = 64; // Buffer size for reading output
var
  rmdirProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
begin
  //debug
  if debug then DebugOut(false,'rmdir','partition',partition);
  //end debug
  setlength(buffer,bufsize);
  termCmd('pkexec rmdir ' + MountFolder + partition,rmdirProcess);
  //output from the command is on stderr
  BytesRead:= rmdirProcess.stderr.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Unable to remove folder ' + MountFolder + partition
                  + LF + buffer,mtError,[mbOK],0)
     end;
  rmdir:= BytesRead;
  rmdirProcess.free;
end; //end function rmdir
{-------------------------------------------------------------------------------
FUNCTION MKDIR
Makes a directory in the default location with the name passed as partition.
Partition will normally be of the form sda1 or sdb3, etc.
The function returns:
  0 = success
 >0 = failure, which will normally be because the folder already exists
      mkdir is used to create a folder to mount a partition to.

Parameters:
  partition - a string containing the partition in the form sda1, sdc3, etc.

Used by:
  Unit utilities - GetPartSize, CheckRoot
  Unit restore - RestGetFile
  Unit clone - CloneFileOpen
  Unit mainform - BackSaveCK

V1.0 - 27 May 2019
V1.1 - 30 Dec 2019 - debug code
V1.2 - 28 Feb 2020 - changed to pkexec mkdir to allow for creating foxclone folder in /media
-------------------------------------------------------------------------------}
function mkdir(partition : string) : integer;
const
  bufsize = 64; // Buffer size for reading output
var
  mkdirProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
begin
  //debug
  if debug then DebugOut(false,'mkdir','partition',partition);
  //end debug
  setlength(buffer,bufsize);
  termCmd('pkexec mkdir ' + MountFolder + partition,mkdirProcess);
  //output from the command is on stderr
  BytesRead:= mkdirProcess.stderr.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Error','Unable to create folder ' + MountFolder + partition
                  + LF + buffer,mtError,[mbOK],0)
     end;
  mkdir:= BytesRead;
  mkdirProcess.free;
end; //end function mkdir
{-------------------------------------------------------------------------------
PROCEDURE PARTED
Reads drive information by running 'parted -l', parsing the output and
populating the global array MyDrives.

parted is used to get drive info:
  model      - e.g. ATA WDC WDS240G2G0A
  mountpt    - e.g. sda
  size       - as a string, e.g. 240GB - may need to change this to UInt64 in KB
  ptabletype - msdos or gpt
  drivetype  - e.g. scsi, sd/mmc

parted is used to get SOME partition info:
  pnumber - e.g. sda1
  mountpt - NO not from parted, this is generated from df output
  psize   - set by parted, but overwritten if available from df
  pused   - NO, not available from parted, set initially = psize and then
            overwritten by df if available (not available from df on unknown fs)
  ptype   - msdos only, e.g. primary, extended, logical
  fs      - e.g. ntfs, ext4, or <unknown>
  pname   - gpt only, set to - for msdos partitions
  plabel  - NO not available from parted, generated by blkid
  flags   - e.g. boot, msftdata

Note, the parted option 'parted -lm' to generate machine readable output
generated garbage output on a FAT32 EFI partition so this procedure does it the
hard way parsing the standard output.

V1.1 - test with encrypted drive installed on MSDOS drive, parted fs = empty but not checking for this.
This check is done for a GPT drive and sets fs=<unknown>. Now also done for MSDOS drive.

V1.2 - significant changes as a result of third party testing - was crashing on a drive with no ptable.
       if statement on partition table type changed to case statement with pchar.
       Was looking to trap a drive with a filesystem with no partition table. Tests on the partition
       table type introduced (parted will report a drive with a valid filesystem but no partition table
       as a 'loop' partition table). Now if a loop partition table type is found a msgbox is used to
       tell the user and the drive is 'excluded' from foxclone - show is set false and then that record
       deleted from the drive array so it will not show in any foxclone window.
       Now how many users are going to use a partition table type other than GPT or MSDOS?

Parameters:
  drives - Tdrives which is an array of Tdrive. This is the array populated by parted
  dcount - an integer, this is populated by parted, the total number of drives found. Could be made
  redundant by replacing all instances with length(MyDrives) - 1.

Used by: Unit mainform - ReadDrives

V1.0 - 27 May 2019
V1.1 - 14 Nov 19
V1.1 - 30 Dec 19 - debug code, also writeln the output of parted
v1.2 - 07 Jan 20
V1.3 - 18 Feb 20 - sudo removed from termCmd('parted -ls',partedProcess) and replaced with pkexec
                   policy added to /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                   to allow parted to run as root.
V1.4 - 21 Mar 20 - had to amend where it was parsing the part #, was reading chars 2-7 and if more than
                   9 parts then missing the '1' on part 10. This was me thinking nobody would be stupid
                   enough to have more than 9 parts on a drive!! Changed to parse chars 1-7.
V1.5 - 14 Apr 20 - not known if an error with the hardware, but both gparted and parted were showing
                   an MS reserved partition on a pcie nvme drive as ext4. It should be unknown so that
                   partclone will use dd to back it up and restore. In foxclone35 this causes the backup
                   to hang. Check added so that if the flag on the partition is msftres then the
                   filesystem is set to <unknown>.
v1.6 - 04 Sep 20 - code tidied on case statement (with statement included) and removed deleting records
                   where show=true - not required doesn't show drives with a loop partition anyway.
                   it is needed without it, it shows the wrong part info on the 2nd box.
                   Now parses a loop part table and includes in the drive array. Msg dialogs on a loop
                   part table removed.
v1.7 - 17 Sep 20 - minor bug where checking for swap at the end was clearing out the ptype, e.g.
                   logical, primary on MSDOS part table so not appearing in partition checkboxes - fixed.
v1.8 - 02 Feb 21 - support removed for reiser and apple filesystems (hfs and hfs+). Now treated as
                   unknown file systems.
-------------------------------------------------------------------------------}
procedure parted(var drives : Tdrives; var dcount : integer);
const
  bufsize = 4096; // Buffer size for reading output
var
  partedProcess  : TProcess;
  partedStream   : TmemoryStream;
  partedData     : TStringList;
  Buffer         : array[1..bufsize] of char;
  BytesRead, i, j,
  dIndex, pIndex : integer;
  sizecol, ptypecol,
  fscol, pnamecol,
  flagscol       : integer; //column numbers for partition information
  pchar : char; //used in case statement on partition table types

begin
  //debug
  if debug then DebugOut(true,'parted');
  //end debug
  termCmd('pkexec parted -ls',partedProcess);
  //-s so it runs without prompts, parted gets confused with some bootable usb
  //drives and throws errors - Ubuntu images (and potentially some other related
  //GNU/Linux distributions) have a peculiar format that allows the image to boot
  //without any further modification from both CDs and USB drives
  partedStream := TMemoryStream.Create;
  //output from partedProcess is read in a loop until no more data is available
  repeat //should get the output in one hit
    BytesRead:= partedProcess.Output.Read(Buffer, bufsize);
    partedStream.Write(Buffer, BytesRead);
  until BytesRead = 0;  // Stop if no more data is available
  partedProcess.Free; //finished with it
  partedStream.Position:= 0; //make sure we start at the beginning of the stream
  partedData:= TStringList.Create;
  partedData.LoadFromStream(partedStream);
  partedStream.Free;
  dIndex:= 0; //index on drives array
  pIndex:= 0; //index on the partitions array
  //debug code
  if debug then
     begin
       for i:= 0 to partedData.count - 1 do writeln(partedData[i]);
       writeln('** end parted output---------------');
     end;
  //end debug

  //now parse the output from parted and put in the drives record array
  for i:= 0 to partedData.count - 1 do // -1 cos indexed from zero
      begin
        //first lines are of the form:
        //Model: ATA ST1000LM048-2E71 (scsi)
        //Disk /dev/sdb: 1000GB
        //Sector size (logical/physical): 512B/4096B - don't care
        //Partition Table: msdos
        //Disk Flags: - don't care

        //get drive model and type
        if pos('Model:',partedData[i]) = 1 then
           begin
             if dIndex <> 0 then //not the first drive block
                begin //set total partitions for drive just parsed and reset pindex
                  drives[dIndex].totalparts:= pindex;
                  pindex:= 0;
                end;
             dIndex:= dIndex + 1;
             setLength(drives, dIndex + 1); //increase the drive array size
             drives[dIndex].show:= true; //make all drives visible
             drives[dIndex].model:= copy(partedData[i],8,99);
             drives[dIndex].model:= trim(copy(drives[dIndex].model,1,pos('(',drives[dIndex].model)-1));
           end;
        //get drive size
        if pos('Disk /',partedData[i]) = 1 then
           begin
             drives[dIndex].mountpt:= copy(partedData[i],11,99); //strip out 'Disk /dev/'
             drives[dIndex].mountpt:= copy(drives[dIndex].mountpt,1,pos(':',drives[dIndex].mountpt)-1);
             drives[dIndex].size:= trim(copy(partedData[i],pos(':',partedData[i])+1,99));
           end;
        //get partition table type
        if pos('Partition Table:',partedData[i]) = 1 then
           drives[dIndex].ptabletype:= uppercase(copy(partedData[i],18,99));
        //parse the data - partition information
        if pos('Number',partedData[i]) = 1 then
           begin
              //get the column numbers for each field used - information is in
              //a fixed format but does vary from disk to disk
              sizecol:= pos('Size',partedData[i]);
              ptypecol:= pos('Type',partedData[i]);
              fscol:= pos('File system',partedData[i]);
              pnamecol:= pos('Name',partedData[i]);
              flagscol:= pos('Flags',partedData[i]);
           end;
         //modified to test for '1' just in case more than 10 partitions
         if (pos(' ',partedData[i]) = 1) or (pos('1',partedData[i]) = 1) then
           //Heading format GPT - Number, Start, End, Size, File-system, Name, Flags
           //Heading format MSDOS - Number, Start, End, Size, Type, File-system, Flags
           //Heading format loop -  Number, Start, End, Size, File-system, Flags
            begin
              pIndex:= pIndex + 1;
              setLength(drives[dIndex].parts, pIndex + 1); //increase the partition array size
              //for a 'std' device, e.g. sda, partitions are sda1, etc. on an SD card mounted through
              //a reader the device is, e,g, mmcblk0 and partitions are mmcblk0p1, p2, etc.
              //nvme will be mounted as nvme0n1 with partitions as nvme0n1p1, etc.
              //if a loop device there will only be ONE partition and it will mount as
              //sda, mmcblk0 or nvme0n1 = same as drive.
              if pos('sd',drives[dindex].mountpt) = 1 then //standard sata drive
                 drives[dIndex].parts[pIndex].pnumber:= drives[dIndex].mountpt
                                                        + Trim(copy(partedData[i],1,7))
              else //its nvme or sd card
                 drives[dIndex].parts[pIndex].pnumber:= drives[dIndex].mountpt
                                                        + 'p' + Trim(copy(partedData[i],1,7));
              //test for loop device and set pnumber to mountpt if true
              if drives[dIndex].ptabletype = 'LOOP' then
                 drives[dIndex].parts[pIndex].pnumber:= drives[dIndex].mountpt;
              drives[dIndex].parts[pIndex].plabel:= '<none>';
              pchar:= drives[dIndex].ptabletype[1]; //get first char so can use with case statement
              case pchar of
                   'G' : with drives[dIndex].parts[pIndex] do //GPT partition table
                              begin
                                //partition type, e.g. primary, extended, etc. n/a for GPT
                                ptype:= '<none>';
                                //file system type, e.g. ntfs, ext4, unknown on ms reserved or extended
                                fs:= Trim(copy(partedData[i],fscol,pnamecol - fscol));
                                //to cope with unknown file systems, e.g. MS reserved partition
                                //this will be later read and partclone.dd used to backup unknown parts
                                if length(fs) = 0 then fs:= '<unknown>';
                                if fs = 'reiserfs' then fs:= '<unknown>';
                                if pos('hfs',fs) = 1 then fs:= '<unknown>';
                                //partition name - GPT only
                                pname:= Trim(copy(partedData[i],pnamecol,flagscol - pnamecol));
                                if length(pname) = 0 then pname:= '<none>';
                                //partition size - in KB
                                psize:= SizeInKB(Trim(copy(partedData[i],sizecol,fscol - sizecol)));
                                //if we have better info later, pused will be overwritten,
                                //but needed for unknown file systems
                                pused:= psize;
                                //partition flags
                                flags:= Trim(copy(partedData[i],flagscol,99));
                                if length(flags) = 0 then flags:= '<none>';
                              end;
                   'M' : with drives[dIndex].parts[pIndex] do //MSDOS partition table
                              begin
                                //partition name - GPT only
                                pname:= '<none>';
                                //partition type, e.g. primary, extended, logical
                                ptype:= Trim(copy(partedData[i],ptypecol,fscol-ptypecol));
                                //file system
                                fs:= Trim(copy(partedData[i],fscol,flagscol - fscol));
                                //to cope with unknown file systems, e.g. MS reserved partition or
                                //encrypted parts this will be later read and partclone.dd used to backup
                                //unknown partitions but also showing <unknown> against extended parts so
                                //modify from the same test for a GPT drive
                                if (length(fs) = 0) and (ptype <> 'extended') then fs:= '<unknown>';
                                //partition size - in KB
                                if fs = 'reiserfs' then fs:= '<unknown>';
                                if pos('hfs',fs) = 1 then fs:= '<unknown>';
                                psize:= SizeInKB(Trim(copy(partedData[i],sizecol,ptypecol - sizecol)));
                                //used space - if MBR drive & extended part set pused=0, else set to psize
                                if ptype = 'extended' then pused:= 0 else pused:= psize;
                                //partition flags
                                flags:= Trim(copy(partedData[i],flagscol,99));
                                if length(flags) = 0 then flags:= '<none>';
                              end;
                    'L' : with drives[dIndex].parts[pIndex] do //loop partition table
                              begin
                                //partition name - GPT only
                                pname:= '<none>';
                                //partition type, e.g. primary, extended
                                ptype:= '<none>';
                                //file system
                                fs:= Trim(copy(partedData[i],fscol,flagscol - fscol));
                                //partition size - in KB
                                psize:= SizeInKB(Trim(copy(partedData[i],sizecol,ptypecol - sizecol)));
                                pused:= psize;
                                //partition flags
                                flags:= Trim(copy(partedData[i],flagscol,99));
                                if length(flags) = 0 then flags:= '<none>';
                              end;
                    else //on case statement
                     with drives[dIndex].parts[pIndex] do
                          begin
                            pnumber:= '';
                            pname:= '<none>';
                            ptype:= '';
                            plabel:= '';
                            fs:= '';
                            OS:= '';
                            psize:= 0;
                            pused:= 0;
                            flags:= '';
                            selected:=false;
                          end;
              end; //case
              //to cater for swap partitions being something like linux-swap(v1), set to swap
              //and if swap then used = 0 as will not be being used by an active linux system
              if pos('swap',drives[dIndex].parts[pIndex].fs) > 0 then
                 with drives[dIndex].parts[pIndex] do
                      begin
                        fs:= 'swap';
                        pused:= 0;
                      end;
           end;
      end; //for i loop - finished parsing parted output
  drives[dIndex].totalparts:= pindex; //for the last block of drive info
  dcount:= length(drives) - 1;
  partedData.Free;
  //check for the MS reserved partition and if the filesystem is not <unknown> then set it.
  //bug in parted showning an MS reserved partition as ext4 on an pcie nvme drive.
  for i:= 1 to dcount do
      for j:= 1 to drives[i].totalparts do
          if pos('msftres',drives[i].parts[j].flags) > 0 then drives[i].parts[j].fs:= '<unknown>';
end; //end procedure parted
{-------------------------------------------------------------------------------
FUNCTION MOUNT
Mounts the filesystem specified by partition (passed as a parameter) in the
default folder. The mount folder needs to be created (mkdir) before calling
mount, otherwise the operation will fail. If the MountFolder is say,
/media/foxclone/ then calling mount('sda3') would mount sda3 as /media/foxclone/sda3.

The function returns:
  0 = success
 >0 = failure, which could be because it is already mounted, or the mount folder
      does not exist

Parameters:
  partition - a string containing the partition to be mounted, e.g. sdb3
  ReadWrite - boolean, if true the partition is mounted read/write, if false
              then the partition is mounted read only.

Used by:
  Unit utilities - GetPartSize, CheckRoot
  Unit restore - RestGetFile
  Unit clone - CloneFileOpen
  Unit mainform - BackSaveCK

V1.0 - 27 May 2019
V1.1 - 30 Dec 2019 - debug code
V1.2 - 18 Feb 2020 - sudo removed from termCmd('sudo mount ' + WriteAccess + '/dev/' + ...
                     and replaced with pkexec, policy defined in
                     /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                     to allow mount to run as root.
-------------------------------------------------------------------------------}
function mount(partition : string; ReadWrite : boolean) : integer;
const
  bufsize = 128; // Buffer size for reading output
var
  mntProcess : TProcess;
  buffer     : string;
  BytesRead  : integer;     //from the stream
  WriteAccess : string;
begin
  //debug
  if debug then
     begin
       if ReadWrite then WriteAccess:= 'true' else WriteAccess:= 'false';
       DebugOut(false,'mount','partition',partition,'ReadWrite',WriteAccess);
     end;
  //end debug
  if ReadWrite then WriteAccess:= '' else WriteAccess:= '-r ';
  setlength(buffer,bufsize);
  termCmd('pkexec mount ' + WriteAccess + '/dev/' + partition + ' ' + MountFolder +partition,mntProcess);
  BytesRead:= mntProcess.output.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Unable to mount /dev/' + partition
                  + LF + buffer,mtError,[mbOK],0)
     end;
  mount:= BytesRead;
  mntProcess.free;
end; //end function mount
{-------------------------------------------------------------------------------
PROCEDURE DFSIZEUSED
Parted does not provide info on the space used on a partition. dfSizeUsed runs
the terminal command df to get both the size of a partition and the space used.
Sizes are reported in 1K blocks. If sdb2 is the partition, dfSizeUsed runs and
parses output from the command "df /dev/sdb2 --output=size,used"
Output parsed is of the form:
  1K-blocks      Used
  328482452 110810208
Note - the partition MUST be mounted
Note - df block sizes are 1024 bytes

Parameters:
  partition - a string containing the partition to get the info on. It is in the
              form sda2, sbc1, etc.
  size - an int64 that is populated with  the size of the partition in blocks
  used - an int64 that is populated with the used space in the partition in blocks

Used by: GetPartSize

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
procedure dfSizeUsed(partition : string; var size :int64; var used : int64);
const
  bufsize = 128; // Buffer size for reading output
var
  dfProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
  sizestr,usedstr : string;
begin
  //debug
  if debug then DebugOut(true,'dfSizeUsed','partition',partition,'size',IntToStr(size),'used',IntToStr(used));
  //end debug
  setlength(buffer,bufsize);
  termCmd('df /dev/' + partition + ' --output=size,used',dfProcess);
  BytesRead:= dfProcess.Output.Read(buffer[1],length(buffer));
  setlength(buffer,BytesRead);
  buffer:= trim(copy(buffer,pos(LF,buffer)+1,length(buffer))); //get rid of the heading
  sizestr:= copy(buffer,1,pos(' ',buffer)-1);
  usedstr:= copy(buffer,pos(' ',buffer)+1,length(buffer));
  size:= StrToInt64(sizestr);
  used:= StrToInt64(usedstr);
  dfProcess.free;
end; //end procedure dfsizeused
{-------------------------------------------------------------------------------
FUNCTION GETLABEL
Returns the label for the specified partition using the command
  blkid /dev/sdb2 | grep LABEL | cut -d= -f2
where sdb2 = partition

Output is of the form = "ultrabay" UUID, which has to be parsed to extract
the label. If there is no label, the function returns an empty string
Labels are only used for information - to help the user identify a partition

Parameters:
  partition - a string containing the partition to get the label on. It is in
              the form sda2, sbc1, etc.

Used by: partLabels

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
V1.2 - 23 Sep 20 - re-running getlabel after a restore does not return the correct labels, remember a
                   post on the LM forum somewhere that running as sudo forces it to re-read the info
                   otherwise it is just reading what is already in memory. Code changed to run with
                   pkexec and a polkit added to allow without password.
                   /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
-------------------------------------------------------------------------------}
function getLabel(partition :string) : string;
const
  bufsize = 128; // Buffer size for reading output
var
  blkidProcess : TProcess;
  buffer     : string;
  BytesRead  : integer;     //from the stream
begin
  //debug
  if debug then DebugOut(false,'getLabel','partition',partition);
  //end debug
  setlength(buffer,bufsize);
  //need label in quotes to get leading space otherwise will also return on partlabel
  termCmd('pkexec blkid /dev/' + partition + ' | grep " LABEL" | cut -d= -f2',blkidProcess);
  BytesRead:= blkidProcess.Output.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       getLabel:=copy(buffer,2,length(buffer));
       getLabel:= copy(getLabel,1,pos('"',getLabel)-1);
     end
  else getLabel:= '<none>';
  blkidProcess.free;
end; //end function getLabel
{-------------------------------------------------------------------------------
PROCEDURE PARTLABELS
Iterates through all partitions in the system and gets partition labels where
available by calling getLabel. Saves outcome in the drives array passed as a
parameter.

Parameters:
  drives -     Tdrives which is an array of Tdrive. This is the array initially
               populated by parted
  maxdrives  - an integer, the total number of drives in the system

Used by: Unit mainform - ReadDrives

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
procedure partLabels(var drives : Tdrives; var maxdrives : integer);
var
  i, j : integer; //loop counters
begin
  //debug
  if debug then DebugOut(true,'partLabels');
  //end debug
  for i:= 1 to maxdrives do //loop through drives
      for j:= 1 to drives[i].totalparts do //loop through partitions
          drives[i].parts[j].plabel:= getLabel(drives[i].parts[j].pnumber);
end; //end procedure partlabels
{-------------------------------------------------------------------------------
FUNCTION UMOUNT
Un-mounts the filesystem specified by partition which would be of the form
sda1, sdb3, etc. Umount does NOT remove the folder where the partition was
mounted, call rmdir to do that.

The function returns:
  0 = success
 >0 = failure, which could be because the partition is wrong or it is not mounted

Parameters:
  partition - a string containing the partition to get the label on. It is in
              the form sda2, sbc1, etc.

Used by:
  Unit utilities - GetPartSize, PTableRestore, CheckRoot
  Unit backup - BackBackup
  Unit restore - RestRestore
  Unit clone - CloneFileToDisk, CloneDiskToDisk
  Unit mainform - BackSaveCK, QuitCK

V1.0 - 27 May 2019
V1.1 - 30 Oct 2019 - debug code
V1.2 - 21 Feb 2020 - sudo removed from termCmd('sudo umount -l ' + '/dev/' + partition,umntProcess);
                     and replaced with pkexec, policy added to
                     /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                     to allow umount to run as root.
-------------------------------------------------------------------------------}
function umount(partition : string) : integer;
const
  bufsize = 128; // Buffer size for reading output
var
  umntProcess : TProcess;
  buffer      : string;
  BytesRead   : integer;     //from the stream
begin
  //debug
  if debug then DebugOut(false,'umount','partition',partition);
  //end debug
  setlength(buffer,bufsize);
  // -l is lazy umount, in testing not all partitions were unmounting
  termCmd('pkexec umount -l ' + '/dev/' + partition,umntProcess);
  //output from the command is on stderr
  BytesRead:= umntProcess.stderr.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Unable to un mount /dev/' + partition
                  + LF + buffer,mtError,[mbOK],0)
     end;
  umount:= BytesRead;
  umntProcess.free;
end; //end function umount
{-------------------------------------------------------------------------------
PROCEDURE - GETPARTSIZE
Reads through all the drives and partitions in MyDrives and gets the partition
size in 1k blocks and used space also in 1k blocks. It uses the procedure
dfSizeUsed to get the size and used size. Must be run AFTER parted otherwise
there will be no drive/partition info in MyDrives. Before calling dfSizeUsed, it
checks to see if a partition is mounted and if not then mounts it. Unknown
partition file systems, extended and swap partitions are ignored.

Parameters:
  drives - Tdrives which is an array of Tdrive. This is the array initially
           populated by parted
  maxdrives - an integer, the total number of drives in the system

Used by: Unit mainform - ReadDrives

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
V1.2 - 05 Jan 20 - modified so that if a drive as been marked as show = false
                   it is ignored
-------------------------------------------------------------------------------}
procedure GetPartSize(var drives : Tdrives; var maxdrives : integer);
var
  i, j : integer; //loop counters
begin
  //debug
  if debug then DebugOut(true,'GetPartSize');
  //end debug
  for i:= 1 to maxdrives do //loop through drives
      if drives[i].show then
         for j:= 1 to drives[i].totalparts do //loop through partitions
             if (drives[i].parts[j].fs <> '<unknown>')
                and (drives[i].parts[j].ptype <> 'extended')
                and (drives[i].parts[j].fs <> 'swap') then
                    begin
                      if CheckMount(drives[i].parts[j].pnumber) = '<none>' then
                         begin
                           if mkdir(drives[i].parts[j].pnumber) = 0 then
                              if mount(drives[i].parts[j].pnumber,false) = 0 then
                                 begin
                                   dfSizeUsed(drives[i].parts[j].pnumber,
                                              drives[i].parts[j].psize,
                                              drives[i].parts[j].pused);
                                   umount(drives[i].parts[j].pnumber);
                                   rmdir(drives[i].parts[j].pnumber);
                                 end
                         end
                      else dfSizeUsed(drives[i].parts[j].pnumber,
                                      drives[i].parts[j].psize,
                                      drives[i].parts[j].pused);
                    end;
end; //end procedure getpartsize

{-------------------------------------------------------------------------------
FUNCTION PARTINFO
Generates a string of partition info for display in partition checkboxes, information is read from a
Tpartparams record.

Order of information (v1.2):
      Mountpoint, e.g. sda1
      file system, e.g. ext4
      [label] if one
      size
      % used
      flags if any
      type, i.e. primary, extended, logical (legacy only) or
      (partition name)
      *OS* if present (and adding a * to the start of partinfo)

Parameters:
  partition - a Tpartparams record containing the partition info to be parsed
  parttype - a string defining the partition table type as either MSDOS or GPT.
             this is required because the information returned by the function
             differs between MSDOS and GPT.

Used by:

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
V1.2 - 19 Sep 20 - changed the order of displayed information
-------------------------------------------------------------------------------}
function PartInfo(var partition : Tpartparams; parttype : string) : string;
const
  ADJUST : real = 1.024; //parted is using 1024 for blocks, df 1000
  MULTIPLY = 1000;
var
  tempnum : real;
begin
  //debug
  if debug then DebugOut(false,'PartInfo','part record for',partition.pnumber,'parttype',parttype);
  //end debug
  with partition do
       begin
         PartInfo:= pnumber + ' : ';
         if ptype = 'extended' then PartInfo+= 'extended ' else
            begin
              PartInfo+= fs;
              if plabel = '<none>' then PartInfo+= ' : : ' else PartInfo+= ' : [' + plabel + '] : ';
              if (fs = '<unknown>') or (fs = 'swap') then tempnum:= psize/MULTIPLY
                 else tempnum:= psize/MULTIPLY * ADJUST;
              if tempnum > 100000 then PartInfo+= FloatToStrF(tempnum/MULTIPLY,ffFixed,4,0) + 'GB'
                 else if tempnum > 1000 then PartInfo+= FloatToStrF(tempnum/MULTIPLY,ffFixed,4,1) + 'GB'
                         else PartInfo+= FloatToStrF(tempnum,ffFixed,4,0) + 'MB';
              PartInfo+= ' : ' + PercentUsed(pused,psize);
              if flags <> '<none>' then PartInfo+= ' : Flags-' + flags + ' : ' else PartInfo+= ' : : ';
              if ptype <> '<none>' then PartInfo+= ptype;
              if pname <> '<none>' then PartInfo+= '(' + pname + ') ';
              //where is the OS
              if OS <> '' then PartInfo:= '*' + PartInfo + ' *' + OS + '*';
            end;


{
         PartInfo+= fs + ' : ';
         if parttype = 'GPT' then PartInfo+= pname + ' : '
            else PartInfo+= ptype + ' : ';
         if ptype <> 'extended' then
            begin
              if (fs = '<unknown>') or (fs = 'swap') then tempnum:= psize/MULTIPLY
                 else tempnum:= psize/MULTIPLY * ADJUST;
              if tempnum > 100000 then PartInfo+= FloatToStrF(tempnum/MULTIPLY,ffFixed,4,0) + 'GB'
                 else if tempnum > 1000 then PartInfo+= FloatToStrF(tempnum/MULTIPLY,ffFixed,4,1) + 'GB'
                         else PartInfo+= FloatToStrF(tempnum,ffFixed,4,0) + 'MB';
              PartInfo+= ' : ' + PercentUsed(pused,psize);
              if plabel <> '<none>' then PartInfo+= ' [' + plabel + ']';
              if flags <> '<none>' then PartInfo+= ' : Flags-' + flags;
            end;
}
       end;
end; //end function partinfo
{-------------------------------------------------------------------------------
FUNCTION PERCENTUSED
Calculates the percentage of two numbers used for display in partition checkbox
information. Returns a string in the form xx.xx% used.

Parameters: a & b both int64, where the result is a/b.

Used by: PartInfo

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
function PercentUsed(a,b : Int64) : string;
var
  answer : real;
begin
  //debug
  if debug then DebugOut(false,'PercentUsed','a',IntToStr(a),'b',IntToStr(b));
  //end debug
  answer:= a/b * 100;
  PercentUsed:= FloatToStrF(answer,ffFixed,3,0) + '% used';
end; //end function percentused
{-------------------------------------------------------------------------------
FUNCTION BLOCKSTOSTR
Takes a number of 1k blocks as reported by df and returns a formated string
of the form 1.1TB, 100GB, 35.6GB, 500MB, 45.6MB

Parameters:
  ablocks - an int64, the number of blocks that have been reported by a previous
            use of df in the terminal.

Used by:
  Unit backup - BackBox4

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
function BlocksToStr(ablocks : int64) : string;
const
  MULTIPLY = 1000;
begin
  //debug
  if debug then DebugOut(false,'BlocksToStr','ablocks',IntToStr(ablocks));
  //end debug
  case ablocks of
       0..100000 : BlocksToStr:= FloatToStrF(ablocks/MULTIPLY,ffFixed,4,1) + 'MB';
       100001..1000000 : BlocksToStr:= FloatToStrF(ablocks/MULTIPLY,ffFixed,4,0) + 'MB';
       1000001..1000000000 : BlocksToStr:= FloatToStrF(ablocks/MULTIPLY/MULTIPLY,ffFixed,4,1) + 'GB';
       1000000001..100000000000 : BlocksToStr:=
                                     FloatToStrF(ablocks/MULTIPLY/MULTIPLY/MULTIPLY,ffFixed,4,1) + 'TB';
  end;
end; //end function blocktostr
{-------------------------------------------------------------------------------
FUNCTION - VALIDPARTTYPE
Was - Generates an error message as a string if a partition is a type that cannot
be saved to. If there is no error, partfailmsg returns an empty string.

V1.2 - function name change and changed from a string fn to a boolean fn. Returns true if the
partition is valid as a destination for a backup, false otherwise. When false generates an
error message.

Parameters:
  partition - a Tpartparams record containing the partition info to be checked

Used by:
  Unit backup - BackBox4

V1.0 - 27 May 2019
V1.1 - 30 Dec 2019 - debug code
V1.2 - 25 Feb 2020
-------------------------------------------------------------------------------}
function ValidPartType(var parts : Tpartparams) : boolean;
const
  cannot = 'Cannot save to';
var
  PartFailMsg : string;
begin
  //debug
  if debug then DebugOut(false,'ValidPartType','part record',parts.pnumber);
  //end debug
  ValidPartType:= false;
  PartFailMsg:= '';
  with parts do
       begin
           if ptype = 'extended' then PartFailMsg:= cannot + LF +'an extended partition';
           if fs = 'swap' then PartFailMsg:= cannot + LF + 'a swap partition';
           if fs = '<unknown>' then PartFailMsg:= cannot + LF + 'an unknown filesystem';
           if pos('esp',flags) > 0 then PartFailMsg:= cannot + LF + 'an EFI partition';
       end;
  if PartFailMsg <> '' then MessageDlg(PartFailMsg,mtWarning,[mbOK],0) else ValidPartType:= true;
end; //end function ValidPartType
{-------------------------------------------------------------------------------
PROCEDURE INITIALSETTINGS
Sets up the settings record with initial default values.

Used by: unit mainform - SettingsDefault, Create - the main form constructor

V1.0 - 27 May 2019
V1.1 - 30 Dec 2019 - debug code
V1.2 - 27 Apr 2020 - amended to set the intial state of fat32
V1.3 - 07 Oct 2020 - default compression changed from 2 to 1.
-------------------------------------------------------------------------------}
procedure InitialSettings;
begin
  //debug
  if debug then DebugOut(true,'InitialSettings');
  //end debug
  with settings do
       begin
         parttable:= true;
         compress:= true;
         complevel:= 1;
         split:= false;
         splitsize:= 2048;
         fat32:= false;
       end;
end; //end procedure initialsettings
{-------------------------------------------------------------------------------
FUNCTION GETPARTCLONEINFO
Used when parsing the output from partclone. Takes the initial output from
partclone as a string and extracts the file system, partition size and used space.
This is returned as a formatted string to display on the progress window (progform)
showing while backing up, restoring and cloning.

Parameters:
  PartCloneStr - the string returned to the calling process containing the initial
                 output from partclone.

Used by: RunPartClone

V1.0 - 27 May 2019
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
function GetPartcloneInfo(var PartCloneStr : string) : string;
var
  tempstr : string;
begin
  //debug
  if debug then DebugOut(false,'GetPartcloneInfo','PartCloneStr',PartCloneStr);
  //end debug
  //get file system, e.g. NTFS
  tempstr:= copy(PartCloneStr,pos('File system:',PartCloneStr),length(PartCloneStr));
  GetPartcloneInfo:= Copy2Symb(tempstr,LF) + LF;
  //get device size
  tempstr:= copy(tempstr,pos('Device size:',tempstr),length(tempstr));
  GetPartcloneInfo+= Copy2Symb(tempstr,'=') + LF;
  //get space in use
  tempstr:= copy(tempstr,pos('Space in use:',tempstr),length(tempstr));
  GetPartcloneInfo+= Copy2Symb(tempstr,'=');
end; //end function getpartcloneinfo
{-------------------------------------------------------------------------------
FUNCTION PARTCLONEB
Builds the command line for running a partclone backup based on the settings in
settings. Returns a command line that can be executed by bash. The form of the
command depends on the settings for compression and splitting files.

Parameters:
  partition - a string of the form sda2, sdb3, etc.
  fs - a string defining the file system for the partition specified, e.g. ntfs
       ext4, etc. This is required because the partclone executable is of the form
       partclone.fs, e.g. partclone.ext4 or partclone.fat32

Used by: Unit backup - BackBackup

V1.0 - 27 May 2019
V1.1 - 23 Nov 2019 - modified to allow for spaces in file/folder names
V1.2 - 30 Dec 2019 - debug code
V1.3 - 22 Feb 2020 - PartCloneB:= 'sudo partclone.'; replaced with PartCloneB:= 'pkexec partclone.';
                     policy added to
                     /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                     to allow partclone to run as root. Note - multiple entries to cover all the
                     variations, e.g. partclone.vfat, partclone.ext4, etc.
-------------------------------------------------------------------------------}
function PartCloneB(partition : string; filesys : string) : string;
var
  LogFile, ImgFile : string;
begin
  //debug
  if debug then DebugOut(false,'PartCloneB','partition',partition,'filesys',filesys);
  //end debug
  PartCloneB:= 'pkexec partclone.';
  //add partclone type and options
  // -c save partition to the special image format (not with partclone.dd)
  // -F force progress (don't know what this means but used in redo)
  // -s source file
  if filesys = '<unknown>' then PartCloneB+= 'dd -F -s /dev/'
  else PartCloneB+= filesys + ' -c -F -s /dev/';
  //add source partition
  PartCloneB+= partition;
  //redirect the partclone log file
  //need quotes round filenames if a space in the filename
  LogFile:= '"' + BackFileName + '.' + partition + '-log.txt"';
  PartCloneB+= ' -L ' + LogFile + ' ';
  //if no compress or split need to add -O = output and overwrite to partclone options
  if not (settings.compress or settings.split) then PartCloneB+= ' -O ';
  //check for and add compression and add pigz
  //where -c write output to stdout and -# is the compression level
  if settings.compress then
     begin
       PartCloneB+= ' | pigz -c -' + IntToStr(settings.complevel);
       if not settings.split then PartCloneB+= ' > ' //redirect pigz output to file
     end;
  //check for and split files where
  // -d is use numeric suffixes starting at 0 for file extn
  // -b put SIZE bytes per output file. The SIZE argument is an integer and
  //    units are K,M,G,T,P,E,Z,Y (powers of 1024) or KB,MB,...(powers of 1000)
  // -  means read standard input
  if settings.split then PartCloneB+= ' | split -d -b ' + IntToStr(settings.splitsize) + 'M - ';
  //Finally add the output file name
  ImgFile:= BackFileName + '.' + partition + '.img';
  if settings.compress then ImgFile+= '.gz';
  if settings.split then ImgFile+= '.';
  //need quotes round filenames if a space in the filename
  ImgFile:= '"' + ImgFile + '"';
  PartCloneB+= ImgFile;
end; //end function partcloneb

{-------------------------------------------------------------------------------
FUNCTION PARTCLONER
Builds the command line for running a partclone restore based the backup filename
(to extract the name out of the *.backup name). The command line is then built
based on settings in RSettings (previously read from the backup file).

Note the used of the -F option to force the restore, partclone was throwing up
an error "Target partition size(0 MB) is smaller than source(20972 MB)" on an
ext4 extended partition (was quite happy with primary partitions in MSDOS &
GPT part tables). Checked with sfdisk and the restored partition table was
NOT showing a zero size partition. Same is also true on the clone procedures.

Test - ferens OS with /home in a logical part on mbr drive. Restored okay with
-F option, but then failed to mount when foxclone run again. Reboot into ferens,
everything okay, its /home is fine. Reboot into LM19 run foxclone, no problems
reading sda and the logical part giving problems????? Not something that would be
encountered normally - after a restore, the next action would be a reboot.

Returns a command line that can be executed by bash.

Parameters:
  RestFname  - is a string and is of the form *.backup and is the FULL pathname
  PartNum - a string of the form sda1 and used to build up the filename for the image file.
  fs - filesystem in the form ext4 or ntfs for the partition to be restored, is used to set the
       version of partclone to run, i.e. partclone.fs = partclone.ntfs or partclone.ext4

               dont like the use of destmountpt it is hiding what is going on and not obvious werhe set

Used by: Unit restore - RestRestore
         Unit clone - CloneFileToDisk

V1.0 - 07 Jun 2019
V1.1 - 23 Nov 2019, modified to allow for spaces in filenames, easiest was to put quotes around the
                    filename irrespective of whether it had spaces in it.
V1.2 - debug code
V1.3 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow partclone to run as root. Note - multiple entries to cover all the
                    variations, e.g. partclone.vfat, partclone.ext4, etc.
V1.4 - 31 Mar 2020, modified so that on a restore the source drive mount point in the backup is replaced
                    by the destination drive mount point. Between the backup and restore the mount point
                    may have changed, e.g. sda to sdx
V1.5 - 14 May 2020, previous versions put quotes around filenames to cater for spaces in filenames. This
                    works if the filename is explicit, but if it contains wildcards, e.g. *, then the
                    cat command will not work. V1.5 removes the quotes and replaces and escapes all
                    spaces in the filename, i.e. ' ' is replaced with '\ '.
V1.6 - 22 Sep 2020, updated on changes to the parameters passed which in turn result from the
                    introduction of the TRestPartData record holding individual info on a partition.
                    Changed the number of params passed, got rid of the optional mount, partstr
                    replaced by partnum.
V1.7 - 19 Jan 2021, modified to generate the log file in DefaultFolder, previously just used the default
                    in partclone which would be overwritten each time it was run.
V1.8 - 04 Feb 2021, cloning from file of sd/nvme to sata is trying to clone to, eg sdap1, fixed
V1.9 - 19 Feb 2021, correcting the bug in the previous version
                    scenarios
                    sata to sata nothing changed
                    sata to sata but destination changed, e.g. sda t sdb
                    nvme to nvme nothing changed
                    nvme to nvme but destination changed - how would this work?
                    nvme to sata
                    sata to nvme
                    partnum is for the source, not the destination
                    destmountpt is the device that partnum needs to be restored to
                    so if partum = sda1 and destmtpt = sdb then partnum needs to be sdb1
                    the code is not pretty
-------------------------------------------------------------------------------}
function PartCloneR(RestFname : string; PartNum : string; fs : string) : string;
var
  Fname : string;
  //added in -F = force cos of problems with extended part.
  POptions : string = ' -r -F -s '; //options are different for partclone.dd, no -r
  LogFname : string;
begin
  //debug
  if debug then DebugOut(false,'PartCloneR','RestFname',RestFname,'Partition',PartNum);
  //end debug
  //get backup filename, add in the partition number and add .img
  Fname:= ReplaceStr(RestFname,'backup',PartNum);
  Fname:= Fname + '.img';
  //escape any spaces in the filename
  if pos(' ',Fname) > 0 then Fname:= StringReplace(Fname,' ','\ ',[rfReplaceAll]);
  //replace the mount point, e.g. sda1 in the backup with the mount point on the destination,
  //e.g. the drive may have moved from sda when the backup was done to sdx.
  Writeln('Partclone R');
  writeln('partnum before changes=',partnum);
  if RSettings.Mountpt = RSettings.DestMountpt then writeln('no changes source/destination') //nothing to change
  else
    begin
       //has sda moved to sdb
      if (pos('sd',RSettings.Mountpt) = 1) and (pos('sd',RSettings.DestMountpt) = 1) then
         begin
           writeln('sata drive has changed mount point from ',RSettings.Mountpt,' to ',RSettings.DestMountpt);
           PartNum:= replacestr(PartNum,RSettings.Mountpt,RSettings.DestMountpt);
         end
      else //source is sata and destination is nvme or vice versa
      if pos('sd',RSettings.Mountpt) = 1 then
         begin
           //destination is nvme or sd card, need to add a p into partnum
           writeln('source was sata, destination is nvme or sd card');
           PartNum:= replacestr(Partnum,RSettings.Mountpt,RSettings.DestMountpt + 'p');
         end
      else //source is nvme and destination is sata, need to take out a p
         begin
           writeln('source was nvme or sd card, destination is sata');
           //strip the p out of the source
           PartNum:= replacestr(PartNum,'p',''); //and then swap source for destination
           PartNum:= replacestr(PartNum,RSettings.Mountpt,RSettings.DestMountpt);
         end;
     end;

  PartNum:= '/dev/' + PartNum;
  writeln('partnum after changes=',partnum);
  //just in case the drive has moved its mount point between the backup and the restore
  //if the destination is nvme then need to add a p

  LogFname:= DefaultFolder + FormatDateTime('yyyymmdd',Now)
             + '.' + ReplaceStr(PartNum,'/dev/','') + '.log';
  writeln('logfilename =',LogFname);
  POptions:= ' -L ' + LogFname + POptions;
  if fs = '<unknown>' then
     begin
       PartCloneR:= 'pkexec partclone.dd';
       POptions:= ' -L ' + LogFname + ' -s ';
     end
  else PartCloneR:= 'pkexec partclone.' + fs;
  //build command line
  with RSettings do
       if Compress and Split then
          // 20190603.sda3.img.gz.01
          // cat 20190603.sda3.img.gz.* | unpigz -c | partclone.fs -d -r -s - -o /dev/sda3
          PartCloneR:= 'cat ' + Fname + '.gz.* | unpigz -c | ' + PartCloneR + POptions + '- -O ' +PartNum
          else if Compress and not Split then
               // 20190604.sda3.img.gz
               // cat 20190603.sda3.img.gz | unpigz -c | partclone.fs -d -r -s - -o /dev/sda3
               PartCloneR:= 'cat ' + Fname + '.gz | unpigz -c | ' + PartCloneR
                            + POptions + '- -O ' + PartNum
               else if not Compress and Split then
                    // 20190604.sda3.img.00
                    // note '-s -' to read input from stream not file
                    // cat 20190603.sda3.img.* | unpigz -c | partclone.fs -d -r -s - -o /dev/sda3
                    PartCloneR:= 'cat ' + Fname + '.* | ' + PartCloneR + POptions + '- -O ' + PartNum
                    else
                      // 20190604.sda3.img
                      // partclone.fs -r -d -s 20190604.sda3.img -o /dev/sda3
                      PartCloneR:= PartCloneR + POptions + Fname + ' -O ' + PartNum;
end; //end function partcloneR
{-------------------------------------------------------------------------------
FUNCTION PARTCLONEC
Builds the command line for running a partclone disk to disk clone.
Returns a command line that can be executed by bash.

Parameters:
  Source - a string of the form /dev/sda1
  Dest - a string of the the form /dev/sdb1
  fs - a string defining the file system, e.g. ext4. This is required because the
       partclone command is of the form partclone.fs, e.g. partclone.ntfs

Used by: Unit clone - CloneDiskToDisk

V1.0 - 02 Jul 2019
V1.1 - 30 Dec 2019 - debug code
V1.3 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow partclone to run as root. Note - multiple entries to cover all the
                    variations, e.g. partclone.vfat, partclone.ext4, etc.
V1.4 - 19 Jan 2021, modified to change logfile name so that unique for each partition copied. User now
                    now has option to view logfiles. Previously the logfile was just overwritten because
                    it was not used for anything.
-------------------------------------------------------------------------------}
function PartCloneC(Source : string; Dest : string; fs : string) : string;
var
  LogFname : string;
begin
  //debug
  if debug then DebugOut(false,'PartCloneC','Source',Source,'Dest',Dest,'fs',fs);
  //end debug
  LogFname:= FormatDateTime('yyyymmdd',Now) + '.' + replacestr(Dest,'/dev/','') + '.log';
  writeln('clone logfilename =',LogFname);
  if fs = '<unknown>' then PartCloneC:= 'pkexec partclone.dd -F --logfile ' + DefaultFolder + LogFname
  else PartCloneC:= 'pkexec partclone.' + fs + ' -F -b --logfile ' + DefaultFolder + LogFname;
  PartCloneC+= ' -s ' + Source;
  PartCloneC+= ' -O ' + Dest;  // -O is overwrite
end; //end function partcloneC
{-------------------------------------------------------------------------------
FUNCTION GETRESTOREDATA
Opens the file specified by filename, reads the contents and populates the
RestSettings record.
The function returns 0 - why???? calling process checks for 0 but does nothing
if anything else and this function always returns 0???

Parameters:
  FileName - full path the the file containing the backup data, of the form
             *.backup.
  Settings - TRestData record a global variable that is populated by the function.

Used by:
  Unit restore - RestGetFile
  Unit clone - CloneFileOpen

V1.0 - 27 May 2019
V1.1 - 21 Oct 19. Swap data read from the backup file
V1.2 - 30 Dec 19, debug code
V1.3 - 09 Jan 20, edited to make the contents of the backup file position insensitive
                  and remove the partition table flag.
V1.4 - 18 Jan 20, error reading restore settings, array index wrong
V1.5 - 12 May 20, strtoint changed to strtoint64
V1.6 - 09 Jun 20, now gets the serial number of the drive in the backup
-------------------------------------------------------------------------------}
procedure getRestoreData(FileName : string; var Settings : TRestData);
const
  Yes = 'YES';
  No = 'NO';
var
  RestData : TStringList; //container for the contents of the *.backup file
  i,j : integer;
  NoteFname : string;
  temp : string;
  debugCompress, debugSplit : string;
begin
  //debug
  if debug then DebugOut(true,'getRestoreData','FileName',FileName);
  //end debug
  RestData:= TStringList.Create;
  RestData.LoadFromFile(FileName);
  Settings.SwapPart:= '';  //zero out
  Settings.SwapUUID:= '';  //zero out
  Settings.BackDate:= RestData[0];
  Settings.PTable:= true;
  NoteFname:= copy(FileName,1,pos('.backup',FileName)-1) + '.note.txt';
  if FileExists(NoteFname) then Settings.Note:= true else Settings.Note:= false;
  j:= 0;
  Settings.Serial:= ''; //foxclone < V40 has no record of the serial number of the drive in the
                        //backup file so set Serial # to empty string
  for i:= 1 to RestData.Count - 1 do
      begin
        if pos('Compression:',RestData[i]) > 0 then Settings.Compress:= copy(RestData[i],13,99) = Yes;
        if pos('Split files:',RestData[i]) > 0 then Settings.Split:= copy(RestData[i],13,99) = Yes;
        if pos('Model:',RestData[i]) > 0 then Settings.Model:= copy(RestData[i],7,99);
        if pos('Serial:',RestData[i]) > 0 then Settings.Serial:= copy(RestData[i],8,99);
        if pos('Mount point:',RestData[i]) > 0 then Settings.Mountpt:= copy(RestData[i],13,99);
        if pos('Partition:',RestData[i]) > 0 then
           begin
             j+= 1;
             setLength(Settings.NewParts, j + 1); //increase the partition array size
             //parse the partition text str into the parts record
             ParsePartData(RestData[i],Settings.NewParts[j])
           end;
        if pos('Blocks:',RestData[i]) > 0 then
           begin
             temp:= copy(RestData[i],8,99); //strips out Blocks to get e.g. sda1,12345
             temp:= copy(temp,1,pos(',',temp)-1); //get the partition, e.g. sda1
             if temp = Settings.NewParts[j].pnumber then
                Settings.NewParts[j].blocks:= StrToInt64(copy(RestData[i],pos(',',RestData[i])+1,99));
           end;
        if pos(': swap :',RestData[i]) > 0 then
           begin
             temp:= copy(RestData[i],11,99); //get rid of Partition text
             Settings.SwapPart:= copy(temp,1,pos(':',temp)-2);
             Settings.SwapUUID:= copy(temp,pos('swap',temp)+7,99);
           end;
      end;
  Settings.TotalParts:= j;
  RestData.free;
  if debug then
     begin
       if Settings.Compress then debugCompress:= Yes else debugCompress:= No;
       if Settings.Split then debugSplit:= Yes else debugSplit:= No;
       DebugOut(true,'getRestoreData','Compress',debugCompress,'Split',debugSplit,
                'Model',Settings.Model,'Mountpt',Settings.Mountpt);
     end;
end; //end function getrestoredata
{-------------------------------------------------------------------------------
PROCEDURE SHOWDRIVES
Populates the checklistbox passed by reference with the contents of the global
record MyDrives. the checklistbox is cleared before it is populated.

Parameters:
  checklistbox - a TChecklistbox passed by reference and populated by the procedure

Used by:
   Unit backup - BackBox2
   Unit clone - CloneFileOpen, CloneDDrive
   Unit mainform - ReadDrives

V1.0 - 27 May 19
V1.2 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
procedure ShowDrives(var checklistbox : TCheckListBox);
var
  drivestr : string;
  i, maxlength : integer;
begin
  //debug
  if debug then DebugOut(true,'ShowDrives','a checklistbox','');
  //end debug
  if checklistbox.Items.Count >= 0 then checklistbox.Clear;
  for i:= 1 to TotalDrives do
      if MyDrives[i].show then
         with MyDrives[i] do
              begin
                drivestr:= mountpt + ' : ' + model + ' : ' + size + ' : ' + 'Type ' + ptabletype;
                checklistbox.Items.Add(drivestr);
              end;
  //turn on scroll bars if needed
  maxlength:=0;
  for i:= 0 to checklistbox.Items.Count - 1 do
      if length(checklistbox.Items.Strings[i]) > maxlength then
         maxlength:= length(checklistbox.Items.Strings[i]);
  if maxlength * 6 > checklistbox.Width then
     checklistbox.ScrollWidth:= checklistbox.Width + 1
  else checklistbox.ScrollWidth:= checklistbox.Width;
  //turn off all check boxes
  for i:= 0 to checklistbox.Count-1 do checklistbox.Checked[i] := False;
end; //end procedure showdrives
{-------------------------------------------------------------------------------
FUNCTION CHECKPARTTABLE
Checks the partition table on a drive and compares it with a partition table
previously created during a backup. The function creates a new partition table
file using the command sfdisk --dump /dev/device and using the command cmd. If a
non zero string is returned by cmd then the partition table has changed.
The function returns:
  false = no change to the partition table
  true = partition table has changed or the current partition table
         cannot be read

Parameters:
  device - a string defining the drive and is of the form sda or sbd, etc.
  fname - the full pathname of the dump file created by the command
          sfdisk --dump during the original backup.

Used by: unit restore - RestBox3

V1.0 - 07 Jun 2019
V1.1 - 24 Nov 19, put quotes around fname in case spaces in the file/folder name
V1.2 - 30 Dec 19, debug code
V1.3 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow sfdisk to run as root.
-------------------------------------------------------------------------------}
function CheckPartTable(device : string; fname : string) : boolean;
const
  bufsize = 64; // Buffer size for reading output
var
  partProcess : TProcess;
  buffer       : string;
  BytesRead    : integer;
begin
  //debug
  if debug then DebugOut(false,'CheckPartTable','device',device,'fname',fname);
  //end debug
    setlength(buffer,bufsize);
    termCmd('pkexec sfdisk --dump /dev/' + device + ' > ' + DefaultFolder + 'temp.sfdisk',partProcess);
    BytesRead:= partProcess.stderr.Read(buffer[1],length(buffer));
    if BytesRead > 0 then
       begin
         setlength(buffer,BytesRead);
         MessageDlg('Unable get current partition table for ' + device
                    + LF + 'the partition table will be restored' + LF + buffer,mtError,[mbOK],0);
         partProcess.free;
         CheckPartTable:= true;
         DeleteFile(DefaultFolder + 'temp.sfdisk');
         exit;
       end;
    partProcess.free;
    fname:= ReplaceStr(fname,'.backup','.sfdisk');
    //put quotes around filename in case of spaces in filename
    fname:= '"' + fname + '"';
    termCmd('cmp ' + DefaultFolder + 'temp.sfdisk ' + fname + ' 2>&1',partProcess);
    //was sure I'd used output before and worked, but not working so changed to stderr
    //wasn't working again, now changed and send stderr to stdout just in case
    BytesRead:= partProcess.output.Read(buffer[1],length(buffer));
    if BytesRead > 0 then
       begin
         CheckPartTable:= true;
         MessageDlg('The partition table has changed since the backup' + LF
                    + 'The partition table will be restored.' + LF
                    + 'Any changes since the backup will be lost.',mtWarning,[mbOK],0);
         partProcess.free;
         DeleteFile(DefaultFolder + 'temp.sfdisk');
         exit;
       end;
    partProcess.free;
    DeleteFile(DefaultFolder + 'temp.sfdisk');
    CheckPartTable:= false;
end; //end function checkparttable
{-------------------------------------------------------------------------------
PROCEDURE RUNPARTCLONE
Runs partclone to either backup or restore a partition and updates the progress
form. Blocks are 1K blocks which will have previously been generated using the
terminal command df.

Parameters:
  cmdstr - the command string to run partclone for either backup or restore
  PartUsedBlks - int64, used blocks in the partition to be backed up or restored,
                 to set the current progress bar maximum
  TotalBlocks - int64, total blocks to backup/restore/clone, used to calculate
                overall % complete and set overall progress bar
  StartTime - time the backup/restore started in datetime format.

Used by:
  Unit backup - BackBackup
  Unit restore - RestRestore
  Unit clone - CloneDiskToDisk, CloneFileToDisk

need to sort out cancelbackup boolean - where is it = in backform also needs
changing

V1.0 - 02 Jul 2019
V1.1 - 30 Dec 19, debug code
V1.3 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow killall to run as root.
V1.4 - 15 Jan 2021, Time replaced with Now in calculating overall time for backup/restore, etc.
                    was used in more than one place.
-------------------------------------------------------------------------------}
Procedure RunPartClone(cmdstr: string; PartUsedBlks : int64;
                       PreviousBlocks : int64;
                       TotalBlocks : int64; StartTime : TDateTime);
const
  bufsize = 2048; // Buffer size for reading output
  ESC = char($1B);
var
  InitialStr : string = ''; //used for initial output from partclone on drive info
  cmdProcess : TProcess; //process to run partclone and read output
  buffer       : string;
  BytesRead    : integer;
  tempstr : string;
  start, finish, nochars : integer;
  CurrentPercent : real;
  OverallPercent : real;
  OverallBlocks : int64;
  ElapsedSecs, RemainSecs : integer;
  blockRate : real;
  RemainTime : TDateTime;
begin
  //debug
  if debug then DebugOut(true,'RunPartClone','cmdstr',cmdstr,'PartUsedBlks',IntToStr(PartUsedBlks),
                         'PreviousBlocks',IntToStr(PreviousBlocks),'TotalBlocks',IntToStr(TotalBlocks),
                         'StartTime',FormatDateTime('hh:nn:ss',StartTime));
  //end debug
  //set current progress bar
  ProgressFm.CurrentProgBar.Max:= PartUsedBlks;
  ProgressFm.CurrentProgBar.Position:= 0;
  //run partclone
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess);
  //read first chunk of output to get partition details
  //note output from partclone on stderr
  repeat
    setlength(buffer,bufsize);
    BytesRead:= cmdProcess.StdErr.Read(buffer[1],length(buffer)); //this works
    setlength(buffer,BytesRead);
    InitialStr+= buffer;
    //searching for ESC because want to read past the first block report from
    //partclone where current block = 0 so avoid div zero later
  until (pos(ESC,Initialstr) > 0);
  if debug then writeln(InitialStr);
  //get device info - the initial output from partclone
  ProgressFm.PartcloneLbl.Caption:= GetPartcloneInfo(Initialstr);
  repeat
    if CancelPartClone then
       begin
         cmdProcess.Terminate(start); //passing any int var as is for windows only
         cmdProcess.Free;
         start:= pos('partclone',cmdstr); //have to pass var not const to ExtractSubstr
         cmdstr:= ExtractSubstr(cmdstr,start,[' ']); //get partclone.xx out of the cmdstr
         cmdstr:= 'pkexec killall ' + cmdstr;
         setlength(buffer,bufsize);
         termCmd(cmdstr,cmdProcess);
         BytesRead:= cmdProcess.StdErr.Read(buffer[1],length(buffer));
         if BytesRead > 0 then
            begin
              setlength(buffer,BytesRead);
              MessageDlg('Error','Failed to kill partclone process' + LF + buffer, mtError,[mbOK],0);
            end;
         cmdProcess.Free;
         ProgressFm.Close;
         MainFm.PageCtrl.Enabled:= true;
         exit;
       end;
    setlength(buffer,bufsize); //2k chunks
    BytesRead:= cmdProcess.StdErr.Read(buffer[1],length(buffer));
    if BytesRead > 0 then
       begin
         setlength(buffer,BytesRead);
         if pos(ESC,buffer) = 0 then //haven't got a complete line so get more
            begin
              tempstr:= buffer;
              BytesRead:= cmdProcess.StdErr.Read(buffer[1],length(buffer));
              if BytesRead > 0 then
                 begin
                   setlength(buffer,BytesRead);
                   buffer:= tempstr + buffer;
                 end;
             end;
         //write(buffer);
         if (pos('Elapsed:',buffer) > 0) and (pos('min,',buffer) > 0) then
            begin
              //The line looks like this for a single partition:
              //Elapsed: 00:03:42, Remaining: 00:00:00, Completed:  99.72%,   5.37GB/min,
              start:= pos('Elapsed:',buffer);
              finish:= pos('min,',buffer);
              nochars:= finish - start + 3; //+3 to get min at the end
              tempstr:= copy(buffer,start,nochars);
              ProgressFm.CurrentProgLbl.Caption:= tempstr;
              //now extract the current % completed
              start:= pos('Completed:',tempstr) + 10;
              finish:= pos('%',tempstr);
              nochars:= finish - start;
              tempstr:= trim(copy(tempstr,start,nochars));
              CurrentPercent:= StrToFloat(tempstr)/100;
              ProgressFm.CurrentProgBar.Position:= round(PartUsedBlks * CurrentPercent);
              OverallPercent:= (PreviousBlocks + round(PartUsedBlks * CurrentPercent))/TotalBlocks;
              OverallBlocks:= round(TotalBlocks * OverallPercent);
              ElapsedSecs:= SecondsBetween(Now,StartTime);
              RemainSecs:= round((TotalBlocks/OverallBlocks - 1) * ElapsedSecs) ;
              RemainTime:= IncSecond(Now,RemainSecs) - Time;
              blockRate:= OverallBlocks/ElapsedSecs / 1024 / 1024 * 60;
              ProgressFm.OverallProgBar.Position:= OverallBlocks;
              ProgressFm.OverallProgLbl.Caption:=
                       'Overall: ' + FormatDateTime('hh:nn:ss',Now - StartTime)
                       + ', Remaining: ' + FormatDateTime('hh:nn:ss',RemainTime)
                       + ', Completed: ' + FloatToStrF(OverallPercent * 100,ffFixed,4,2) + '%,  '
                       + FloatToStrF(blockRate,ffFixed,4,2) + 'GB/min';
            end;
         if (pos('Syncing',buffer) > 0) then ProgressFm.CurrentProgLbl.Caption:= 'Syncing...';
         if (pos('successfully',buffer) > 0) then
            ProgressFm.CurrentProgLbl.Caption:= 'Cloned successfully';
         application.processmessages;
       end;
  until BytesRead = 0;
  cmdProcess.Free;
end; //end procedure RunPartClone
{-------------------------------------------------------------------------------
FUNCTION DRIVESIZE
Returns the size of a drive in sectors, where sector = 512B by running the
terminal command blockdev.

Parameters:
  Drive - a string defining the drive as sda, sdd, etc.

Used by: ModifyPTable
         Unit clone - CloneCheck

V1.0 - 02 Jul 2019
V1.1 - 30 Dec 2019, debug code
V1.2 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow blockdev to run as root.
V1.2 - 11 May 2020, strtoint changed to strtoint64 because it was reporting the wrong size for a
                    2TB drive.
-------------------------------------------------------------------------------}
function DriveSize(Drive : string) : int64;
const
  bufsize = 64;
var
  cmdstr : string;
  buffer : string;
  cmdProcess : TProcess;
  BytesRead : integer;
begin
  //debug
  if debug then DebugOut(false,'DriveSize','Drive',Drive);
  //end debug
  cmdstr:= 'pkexec blockdev --getsz /dev/' + Drive;
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess); //run blockdev
  BytesRead:= cmdProcess.Output.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       buffer:= copy(buffer,1,length(buffer)-1); //get rid of the LF at end
       DriveSize:= StrToInt64(buffer);
     end
  else DriveSize:= 0;
  cmdProcess.free;
end; //end function DriveSize
{-------------------------------------------------------------------------------
PROCEDURE GRUBRESTORE
Restores GRUB from either a file or another device. boot.img will be restored
by restoring the partition table, the main purpose of grubrestore is to restore
core.img (the second part of the bootloader) which sits in the gap between the
end of the partition table and the start of the first partition. This is
assumes to start at 1MB. GrubRestore restores the first 1048576 bytes of the
disk, i.e. 1MB.

V1.3 - changed, no longer restores the first 1MiB of the drive as GrubBackup only
saves the space between the start of the drive and the start of the first partition (less one block).
The first partition may not start at 1MiB (=2048 blocks). Now just restores the contents of the
.grub file.

See https://en.wikipedia.org/wiki/GNU_GRUB for more info on grub.

Parameters:
  Source - string defining the source as either a drive, e.g. sda or a file
           containing a previous dump of the first MB of a drive.
  Target - string defining the target drive, e.g. sdb.
  FileFlag - boolean, when true the source is a file inc. full path name, when
             false, the source is a drive.

Used by: unit restore - RestRestore
         unit clone - CloneDiskToDisk, CloneFileToDisk

V1.0 - 13 Oct 19
V1.1 - 24 Nov 19, quotes around the filename to cater for spaces in the filename
V1.2 - 30 Dec 19, debug code
V1.3 - 07 Jan 20, modified to simplify the dd command to dd if=xx of=yy
V1.4 - 06 Feb 20, because of the changes in V1.3, when cloning a drive it copies the whole drive,
                  e.g. dd if=/dev/sda of=/dev/sdb,
                  amended to use grubSize to determine how many 512B blocks to copy
V1.5 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow dd to run as root.
-------------------------------------------------------------------------------}
procedure GrubRestore(Source : string; Target : String; FileFlag : boolean);
const
  bufsize = 64; // Buffer size for reading output
var
  grubstr, buffer, filestr : string;
  grubProcess : Tprocess;
  BytesRead : integer;
begin
  //debug
  if debug then
     begin
       if FileFlag then filestr:= 'true' else filestr:= 'false';
       DebugOut(true,'GrubRestore','Source',Source,'Target',Target,'FileFlag',filestr);
     end;
  //end debug
  //quotes around the filename in case of spaces in the filename
  if FileFlag then grubstr:= 'pkexec dd if="' + Source + '"' + ' of=/dev/' + Target
  else grubstr:= 'pkexec dd if=/dev/' + Source + ' of=/dev/' + Target + ' count=' + grubSize(Source);
  setlength(buffer,bufsize);
  termCmd(grubstr,grubProcess); //run dd
  //output from the command is on stderr but dd still generates output when successful so ignore
  BytesRead:= grubProcess.stderr.Read(buffer[1],length(buffer));
  grubProcess.free;
end; //end procedure GrubRestore
{-------------------------------------------------------------------------------
FUNCTION PTABLERESTORE
Restores a partition table. Had initial problems with partclone reporting a
zero length logical partition. Was not reading the output from sfdisk so assume
that it was not completing before partclone was running. Changed to read output and
changed from procedure to a function.

V1.1 - thought that sfdisk updated the kernel on completion, but obviously not. Deleting a swap
partition and then restoring it with mkswap was throwing up errors - cannot open /dev/sdxx. Gparted was
also showing errors - unable to probe device, no such file or directory.
partprobe added to the function to inform the OS of changes. Don't understand why it worked okay on
restoring 'ordinary' partitions?

V1.2 - quotes added around filenames to cater for spaces in filename or folder

V1.3 - given problems cloning disks (was using file to disk) - worked but showed errors in gparted after,
make sure all partitions on the target drive are unmounted before restoring the part table. Error from
sfdisk was:

    Checking that no-one is using this disk right now ... FAIL

and at the time there was at least one partition on the drive that was mounted.

V1.4 - working fine running under LM19.0 but not under ubuntu on the stick, changed code so that if
partprobe returns an error then the function exits with an error.

V1.5 - sorted, problem was systemd automatically using any swap parts it finds on boot. swapoff in
rc.local. revert to earlier version

Returns true if successful and false otherwise.

Parameters:
  Source - string defining a partition table file previously created by sfdisk.
           It is the full path name.
  Target - string defining the target drive, e.g. sdb.

Used by: unit clone - CloneDiskToDisk, CloneFileToDisk
         unit restore - RestRestore

V1.0 - 13 Oct 2019
V1.1 - 29 Oct 2019
V1.2 - 25 Nov 2019
V1.3 - 25 Nov 2019
V1.5 - 03 Dec 2019
V1.6 - 30 Dec 2019, debug code
V1.5 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow sfdisk and partprobe to run as root.
-------------------------------------------------------------------------------}
function PTableRestore(Source : string; Target : string) : boolean;
const
  bufsize = 4096; // sfdisk can generate quite a lot of output
var
  cmdstr, buffer : string;
  cmdProcess : TProcess;
  BytesRead : integer;
  sfdiskstr : string = '';
  i,j : integer;
begin
  //debug
  if debug then DebugOut(false,'PTableRestore','Source',Source,'Target',Target);
  //end debug
  //check and unmount any mounted partitions on the target
  for i:= 1 to TotalDrives do if MyDrives[i].mountpt = Target then
     begin
       if debug then writeln('** target found, unmounting partitions on ',Target);
       for j:= 1 to MyDrives[i].totalparts do if CheckMount(MyDrives[i].parts[j].pnumber) <> '<none>'
        then
          begin
            if debug then writeln('** This partition was mounted = ', MyDrives[i].parts[j].pnumber);
            umount(MyDrives[i].parts[j].pnumber);
          end;
      end;
  //create partition table on target drive
  //--force to make it do it even if it thinks something is mounted
  cmdstr:= 'pkexec sfdisk --force /dev/' + Target + ' < "' + Source + '"';
  termCmd(cmdstr,cmdProcess); //run sfdisk to restore part table
  repeat
    setlength(buffer,bufsize);
    BytesRead:= cmdProcess.Output.Read(buffer[1],length(buffer)); //this works
    setlength(buffer,BytesRead);
    sfdiskstr+= buffer;
  until BytesRead = 0;
  //debug
  if debug then writeln(sfdiskstr);
  //now partprobe to get the kernel to reread the part table
  cmdstr:= 'pkexec partprobe /dev/' + Target;
  termCmd(cmdstr,cmdProcess);
  setlength(buffer,bufsize);
  BytesRead:= cmdProcess.stderr.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       MessageDlg('Partition table error','Failed to restore partition table to ' + Target
                   + LF + 'Restore/clone cannot continue'
                   + LF + buffer, mtError,[mbOK],0);
       writeln(Buffer);
       PTableRestore:= false;
     end
  else PTableRestore:= true;
  cmdProcess.free;
end; //end function PTableRestore
{-------------------------------------------------------------------------------
FUNCTION MODIFYPTABLE
Changes a partition table file previously created by sfdisk to point at a new
drive and change its size. If the source drive was sda and the target is sdb,
it changes all occurrences of sda with sdb in the file.
It reads the size of the target drive with drivesize and modifies the 'last-lba:'
line in the file to reflect the target drive size. The number of blocks is
reduced by 34 to reflect that fact that on a GPT drive there is a backup copy of
the partition table at the end of the drive in the last 34 blocks.
Note - this only applies to a partition table file from a GPT format drive,
n/a on an MSDOS partition table.
The output can be saved to a different file or the original overwritten depending
on the parameter passed as outfile.

The function returns true if successful, false means it failed to get the size
of the target drive.

Parameters:
  InFile - a string containing the full pathname of the sfdisk partition table file.
  OutFile - a string defining the full pathname for saving the modified partition
            table file.
  Target - the target drive that the partition table file will be amended for,
           in the form sda, sdb, etc.

Used by: unit clone - CloneFileToDisk, CloneDiskToDisk

V1.0 - 29 Jul 2019
V1.1 - 25 Nov 2019 - quotes around filename to cater for spaces
V1.2 - 30 Dec 2019 - debug code
V1.3 - 05 Mar 2020 - bug, was not translating an nvme/sd card correctly.
V1.4 - 04 Feb 2021 - bug, cloning sd/nvme to sata still leaving the 'p' in. fixed.
-------------------------------------------------------------------------------}
function ModifyPTable(InFile : string; OutFile : string; Target : string) : boolean;
var
  i : integer;
  sfdiskData : TStringList;
  lastBlock : int64 = 0; //last block on the target drive
  Source : string; //source drive in the drive in infile
begin
  //debug
  if debug then DebugOut(false,'ModifyPTable','InFile',InFile,'OutFile',OutFile,'Target',Target);
  //end debug
  lastBlock:= DriveSize(Target);
  if lastBlock = 0 then
     begin
       MessageDlg('Drive error','Failed to get size of drive ' + Target
                  + LF + 'Clone cannot continue', mtError,[mbOK],0);
       ModifyPTable:= false;
       Exit;
     end;
  sfdiskData:= TStringList.Create;
  sfdiskData.LoadFromFile(InFile);
  Source:= ReplaceStr(sfdiskData[2],'device: /dev/','');
  //if source is a sata drive and target is an NVME drive, partitions on the NVME
  //drive are drive + p1, p2, etc.
  if pos('sd',Target) <> 1 then Target:= Target + 'p';
  //also need to check if nvme/sd to nvme/sd dont want to add a p if it is already there
  for i:= 0 to sfdiskData.count - 1 do
      begin
        if pos(Source,sfdiskData[i]) > 0 then sfdiskData[i]:= ReplaceStr(sfdiskData[i],Source,Target);
        //need to check if source is nvme/sd and dest is sata, need to strip out the p from the
        //part number, e.g. /dev/mmcblk0p1 to sda1 not sdap1
        //gpt drive, make space at the end of the drive for the backup partition table
        if pos('last-lba:',sfdiskData[i]) > 0 then sfdiskData[i]:= 'last-lba: ' + IntToStr(lastBlock-34);
        if pos('type=5',sfdiskData[i]) > 0 then
           begin
            //msdos drive and found extended partition, would expect it to use up the
            //whole drive, but it might not. Need to check where it ends and set it
            //to the target size if target is smaller.
           end;
      end;
  //on an nvme drive or sd card this will have placed a 'p' at the end of the line[2] device: e.g.
  //      device: /dev/mmcblk0p
  //this should be device: /dev/mmcblk0, so strip out the p. Work on the basis that unlikely to have
  //16 sata devices on a system so just check last char = p.
  if copy(sfdiskData[2],length(sfdiskData[2]),1) = 'p' then
     sfdiskData[2]:= copy(sfdiskData[2],1, length(sfdiskData[2])-1);
  //if the source is nvme/sd and destination is sata then partitions will now be named, e.g. sdap1
  //need to strip out the p
  if (pos('sd',Source) = 0) and (pos('sd',Target) <> 0) then
     for i:= 0 to sfDiskData.count-1 do //strip out the p from /dev/sdap1
         if pos('/dev/',sfDiskData[i]) = 1 then sfDiskData[i]:= replacestr(sfDiskData[i],Target + 'p',Target);
  sfdiskData.SaveToFile(OutFile);
  sfdiskData.Free;
  ModifyPTable:= true;
end; //end function ModifyPTable

{-------------------------------------------------------------------------------
FUNCTION GETUUID
Returns a string containing the UUID for the specified device. If a UUID is not
found for the specified partition the function returns an empty string.

Parameters:
  Partition - a string in the form sdax or equivalent for the partition that the
  UUID is required for.

Used by: unit backup - BackBackup
         unit clone - CloneDiskToDisk

V1.0 - 15 Oct 2019
V1.1 - 30 Dec 2019 - debug code
V1.2 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow blkid to run as root.
-------------------------------------------------------------------------------}
function getUUID(partition : string) : string;
const
  bufsize = 64;
var
  cmdstr, buffer : string;
  cmdProcess : TProcess;
  BytesRead : integer;
begin
  //debug
  if debug then DebugOut(false,'getUUID','partition',partition);
  //end debug
  getUUID:= '';
  cmdstr:= 'pkexec blkid /dev/' + partition + ' | cut -f2 -d\"';
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess);
  BytesRead:= cmdProcess.Output.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       getUUID:= copy(buffer,1,length(buffer)-1); //get rid of the LF at end;
     end;
  cmdProcess.free;
end; //end function getUUID
{-------------------------------------------------------------------------------
PROCEDURE RESTORESWAP
Restores a swap partition

Parameters:
  Partition - a string in the form sdax or equivalent for the partition that the UUID is required for.
  UUID - a string containing the UUID for the swap partition to be restored

Used by: unit restore - RestRestore
         unit clone - CloneFileToDisk, CloneDiskToDisk

V1.0 - 28 Oct 2019
V1.1 - 30 Dec 2019 - debug code
V1.2 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow mkswap to run as root.
-------------------------------------------------------------------------------}
procedure RestoreSwap(partition : string; UUID : string);
const
  bufsize = 128;
var
  cmdstr, buffer, swapstr : string;
  cmdProcess : TProcess;
  BytesRead : integer;
begin
  //debug
  if debug then DebugOut(true,'RestoreSwap','partition',partition,'UUID',UUID);
  //end debug
  //CmdLine =pkexec mkswap -U d785dbd8-2558-4993-ac1d-ba09202f60d0 /dev/mmcblk07
  cmdstr:= 'pkexec mkswap -U ' + UUID + ' /dev/' + partition;
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess);
  BytesRead:= cmdProcess.stderr.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       swapstr:= copy(buffer,1,length(buffer));
     end;
  //mkswap outputs on failure include "error" and "cannot", it is not a terminal failure which is why
  //RestoreSwap is a procedure not a boolean function. But generate a warning message if encountered
  if (pos('error',swapstr) > 0) or (pos('cannot',swapstr) > 0) then
          MessageDlg('Error restoring the swap partition ' + LF + swapstr,mtWarning,[mbOK],0);
  cmdProcess.free;
end; //end procedure RestoreSwap
{-------------------------------------------------------------------------------
FUNCTION CHOPSTR
Chops a string to remove the leading chars until the string is of length less than specified by chop.

Used to chop strings in some tlabels where turning on wordwrap doesn't work cos there are
no spaces in the string to split at. Chops the string at '/' to progressively remove the leading
parts of the full pathname.

Parameters:
  Astring - the string to be chopped
  Chop - integer, maximum length of the string to return

Used by: unit backup - BackConfirm
         unit clone - CloneFileConfirm
         unit restore - RestConfirm

V1.0 - 4 Nov 19
V1.1 - 25 Nov 19, modified to strip out the initial mount point, e.g if the string is
                  /media/Ultrabay/MBRfull7bodhi/somefile.txt, it gives MBRfull7bodhi/somefile.txt
V1.2 - 30 Dec 19 - debug code
V1.3 - 04 Sep 20 - modified to chop out the leading chars from the string at "/" char
V1.4 - 17 Sep 20 - got the test on the while loop wrong, when into infinite loop. Added StartStr so that
                   .../ wasn't prepended to an unchanged Astring.
-------------------------------------------------------------------------------}
function ChopStr(AString : string; Chop : integer) : string;
var
  Startstr : string;
begin
  //debug
  if debug then DebugOut(false,'ChopStr','AString',AString,'Chop',IntToStr(Chop));
  //end debug
  Startstr:= Astring;
  while (length(Astring) > (Chop - 4)) and (pos('/',astring) <> 0) do
          if pos('/',Astring) > 0 then Astring:= copy(Astring,pos('/',Astring)+1,99);
  if StartStr <> AString then ChopStr:= '.../' + Astring else Chopstr:= Startstr;
end; //end function ChopStr
{-------------------------------------------------------------------------------
FUNCTION GETVERSION
Returns a string that contains the version number plus kernel version plus date where
version and versiondate are defined global constants. The kernel version is determined
using 'uname -r'.

Why - so that the current kernel is displayed in the version label so that users
can see what kernel is being used if they have problems with specific kernels.

Used by: unit mainform - Create - used by VersionLbl & it is displayed on the main window bottom right
         unit mainform - CreateAboutPage

V1.0 - 6 Nov 19
V1.1 - 30 Dec 19 - debug code
-------------------------------------------------------------------------------}
function GetVersion : string;
const
  bufsize = 32;
var
  cmdstr, buffer : string;
  cmdProcess : TProcess;
  BytesRead : integer;
begin
  //debug
  if debug then DebugOut(false,'GetVersion');
  //end debug
  cmdstr:= 'uname -r ';
  setlength(buffer,bufsize);
  termCmd(cmdstr,cmdProcess);
  BytesRead:= cmdProcess.Output.Read(buffer[1],length(buffer));
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       GetVersion:= copy(buffer,1,length(buffer));
     end;
  if pos('-generic',GetVersion) > 0 then GetVersion:= copy(GetVersion,1,pos('-generic',GetVersion)-1);
  GetVersion:= Version + ' k' + GetVersion + ', ' + VersionDate;
  cmdProcess.free;
end; //end function GetVersion
{-------------------------------------------------------------------------------
PROCEDURE OSPROBE
Uses os-prober to determine what OSes are present in the system and populates
mydrives with the result.

Used by: unit mainform - ReadDrives

V1.0 - 07 Nov 2019
V1.1 - 30 Dec 2019, debug code
V1.2 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow os-prober to run as root.
-------------------------------------------------------------------------------}
procedure OSprobe;
const
  bufsize = 512; // Buffer size for reading output
var
  OSprocess  : TProcess;
  OSstream   : TmemoryStream;
  OSdata     : TStringList;
  Buffer         : array[1..bufsize] of char;
  BytesRead, i, j, k : integer;
  OSpart, OStitle : string;
begin
  //debug
  if debug then DebugOut(false,'OSprobe');
  //end debug
  termCmd('pkexec os-prober',OSprocess);
  OSstream := TMemoryStream.Create;
  repeat //should get the output in one hit
    BytesRead:= OSprocess.Output.Read(Buffer, bufsize);
    OSstream.Write(Buffer, BytesRead);
  until BytesRead = 0;  // Stop if no more data is available
  OSprocess.Free; //finished with it
  OSstream.Position:= 0; //make sure we start at the beginning of the stream
  OSdata:= TStringList.Create;
  OSdata.LoadFromStream(OSstream);
  OSstream.Free;
  //set all OS entries for all partitions to an empty string
  for j:= 1 to TotalDrives do for k:= 1 to MyDrives[j].totalparts do MyDrives[j].parts[k].OS:= '';
  for i:= 0 to OSdata.count - 1 do
      begin
        OSpart:= copy(OSdata[i],6,pos(':',OSdata[i])-6);
        OStitle:= copy(OSdata[i],pos(':',OSdata[i])+1,99);
        OStitle:= copy(OStitle,1,pos(':',OStitle)-1);
        if pos('(',OStitle) > 0 then OStitle:= trim(copy(OStitle,1,pos('(',OStitle)-1));
        //add the OStitle into the record for the relevant partition
        for j:= 1 to TotalDrives do
            for k:= 1 to MyDrives[j].totalparts do
                if MyDrives[j].parts[k].pnumber = OSpart then MyDrives[j].parts[k].OS:= OStitle;
      end;
  OSdata.free;
end; //end procedure OSprobe

{-------------------------------------------------------------------------------
PROCEDURE DEBUGOUT
Debugout dumps the function/procedure name to terminal together with the parameters passed to it.
In foxcloneXX.pas, any options passed in the command line are read and if it is -d or --dump then the
debug flag is set. All utility functions/procedures check this and if true use debugout to dump
information to the terminal.

Testing identified bugs causing foxclone to crash which led to this procedure being included. It helps
identify where in the code foxclone crashed and what parameter values might have caused it to crash.

Parameters:
  proc - boolean, if true procedure is sent to the terminal, false is a function
  fname - string, the name of the calling function/procedure
  pX - string, the name of the parameter passed to the function/procedure
  pXtext - string, the value of the parameter pX. Note if not a string it must be converted to a string
           before passing it to debugout.

The procedure allows for 6 parameter/value pairs.

Used by:  too many to list

V1.0 - 4 Jan 2020
-------------------------------------------------------------------------------}
procedure DebugOut(proc : boolean; fname : string;
                   p1 : string = ''; p1text : string = '';
                   p2 : string = ''; p2text : string = '';
                   p3 : string = ''; p3text : string = '';
                   p4 : string = ''; p4text : string = '';
                   p5 : string = ''; p5text : string = '';
                   p6 : string = ''; p6text : string = '');
begin
  if proc then write('** procedure ') else write('** function ');
  writeln(fname);
  if p1 <> '' then writeln('** ',p1,' =',p1text);
  if p2 <> '' then writeln('** ',p2,' =',p2text);
  if p3 <> '' then writeln('** ',p3,' =',p3text);
  if p4 <> '' then writeln('** ',p4,' =',p4text);
  if p5 <> '' then writeln('** ',p5,' =',p5text);
  if p6 <> '' then writeln('** ',p6,' =',p6text);
  writeln('** ---------------');
end; //procedure DebugOut

{-------------------------------------------------------------------------------
PROCEDURE GRUBBACKUP
Backs up grub to a file. The procedure runs sfdisk --dump to find out where the first partition on the
drive starts and then runs dd to copy the first xx 512byte blocks of the drive where xx is the start of
the first partition minus one.

Previously foxclone just copied the first 1MiB of the drive, but the first partition might not start at
the 1MiB boundary.

Parameters:
  MountPoint - string, e.g. sda
  Filename - full pathname for the backup file set, GrubBackup appends .grub to the filename and saves
  grub as the filename.

Used by:  unit backup - BackBackup

V1.0 - 07 Jan 2020
V1.1 - 06 Feb 2020, modified to use a separate function grubSize to determine grub size
V1.2 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow dd to run as root.
V1.3 - 22 Aug 201, spaces in filename cause problems and not previously spotted = no grub file, now
                   escapes the filename
-------------------------------------------------------------------------------}
procedure GrubBackup(MountPoint : string; Filename : string);
const
  bufsize = 2048; // Buffer size for reading output
var
  cmdstr : string;
  grubStream   : TMemoryStream;
  BytesRead    : integer;
  grubProcess  : TProcess;
  Buffer       : array[1..bufsize] of char;
begin
  //debug
  if debug then DebugOut(true,'GrubBackup','MountPoint',MountPoint,'Filename',Filename);
  //end debug
  //escape any spaces in the filename
  Filename:= replacestr(Filename,' ','\ ');
  //now run dd to copy grub to file
  cmdstr:= 'pkexec dd if=/dev/' + MountPoint + ' of=' + Filename + '.grub ' + 'count=' + grubSize(MountPoint);
  grubStream:= TMemoryStream.Create;
  termCmd(cmdstr,grubProcess); //run dd
  //output from the command is on stderr but dd still generates output when successful so ignore
  repeat //should get the output in one hit
    BytesRead:= grubProcess.stderr.Read(Buffer, bufsize);
    grubStream.Write(Buffer, BytesRead);
  until BytesRead = 0;  // Stop if no more data is available
  grubStream.free;
  grubProcess.free;
end; //procedure GrubBackup
{-------------------------------------------------------------------------------
FUNCTION GRUBSIZE
Previous versions of foxclone assumed that the first partition started at 1MB and when copying grub
simply copied the first 1MB of the drive. This function returns the start of the first partition minus 1
in 512B blocks for use by the grub backup and restore procedures.

Returned as a string so that the output can be used directly in a dd command.

Parameters:
  Drive - string, e.g. sda

Used by:  GrubBackup
          GrubRestore

V1.0 - 06 Feb 2020
V1.1 - 22 Feb 2020, sudo replaced with pkexec and policy added to
                    /usr/share/polkit-1/actions/com.foxclone.foxclone.policy
                    to allow sfdisk to run as root.
V1.2 - 22 Aug 2021, Start was an integer, which the last partition starting at around 1TB this resulted
                    Start returning a negative number = no grub file saved. Changed to int64.
-------------------------------------------------------------------------------}
function grubSize(Drive : string) : string;
const
  bufsize = 2048; // Buffer size for reading output
var
  cmdstr : string;
  grubStream   : TMemoryStream;
  BytesRead, i : integer;
  grubProcess  : TProcess;
  Buffer       : array[1..bufsize] of char;
  grubData     : TStringList;
  Startstr : string;
  Start : int64 = 2147483647; //max for signed 32 bit int - first sector of first partition
begin
  //debug
  if debug then DebugOut(false,'grubSize','Drive',Drive);
  //end debug
  cmdstr:= 'pkexec sfdisk --dump /dev/' + Drive;
  grubStream:= TMemoryStream.Create;
  termCmd(cmdstr,grubProcess);
  repeat //should get the output in one hit
    BytesRead:= grubProcess.Output.Read(Buffer, bufsize);
    grubStream.Write(Buffer, BytesRead);
  until BytesRead = 0;  // Stop if no more data is available
  grubProcess.free; //finished with it
  grubStream.Position:= 0; //make sure we start at the beginning of the stream
  grubData:= TStringList.Create;
  grubData.LoadFromStream(grubStream);
  grubStream.free;
  for i:= 0 to grubData.count - 1 do
       if pos('start=',grubData[i]) > 0 then
          begin
            Startstr:= copy(grubData[i],pos('start=',grubData[i])+7,99);
            Startstr:=trim(copy(startstr,1,pos(',',startstr)-1));
            //changed from StrToInt to StrToInt64
            if StrToInt64(Startstr) < Start then Start:= StrToInt64(Startstr);
          end;
  grubData.free;
  grubSize:= IntToStr(Start - 1);
end; //function grubSize
{-------------------------------------------------------------------------------
FUNCTION - FREESPACE
Was - Generates an error message as a string if a partition is a type that cannot
be saved to. If there is no error, partfailmsg returns an empty string.

V1.2 - function name change and changed from a string fn to a boolean fn. Returns true if the
partition is valid as a destination for a backup, false otherwise. When false generates an
error message.

Parameters:
  partition - a Tpartparams record containing the partition info to be checked

Used by:
  Unit backup - BackBox4

V1.0 - 25 Feb 2020
-------------------------------------------------------------------------------}
function FreeSpace(var parts : Tpartparams) : boolean;
var
  i : integer;
  TotalBlocks : int64 = 0; //total 1K blocks used in the partitions to backup
  BlocksFree : int64;      //free 1K blocks on the destination partition
begin
  //debug
  if debug then DebugOut(false,'FreeSpace','part record',parts.pnumber);
  //end debug
  for i:= 1 to BackDrive.totalparts do
      if BackDrive.parts[i].selected then TotalBlocks+= BackDrive.parts[i].pused;
  BlocksFree:= parts.psize - parts.pused - TotalBlocks;
  if BlocksFree < 0 then //not enough space
     begin
       MessageDlg('Not enough space on destination' + LF
                  + 'Space required = ' + #9 + BlocksToStr(TotalBlocks) + LF
                  + 'Space available = ' + #9 + BlocksToStr(parts.psize - parts.pused)
                  + LF + 'Choose another drive/partition', mtWarning, [mbOK], 0);
       FreeSpace:= false;
     end
  else FreeSpace:= true;
end; //end function FreeSpace
{-------------------------------------------------------------------------------
PROCEDURE GETSERIALS
Gets the serial numbers for the drives in the system. Previously relied on the model name for the
unique identifier for a drive. But if there are two identical drives in a system???

Reads through the drives in MyDrives and gets the serial number for each drive.
Must be run after parted has populated the MyDrives record.

Parameters:
  drives - Tdrives which is an array of Tdrive. This is the array populated by parted
  dcount - an integer, this is populated by parted, the total number of drives found. Could be made
  redundant by replacing all instances with length(MyDrives) - 1.

Used by:  unit mainform - TMainFm.ReadDrives

V1.0 - 09 June 2020
V1.1 - 12 June 2020, amended to check for a CD drive as foxclone was throwing up an error - couldn't get
                     serial # on a CD drive.
-------------------------------------------------------------------------------}
procedure getSerials(var drives : Tdrives; var dcount : integer);
var
  i : integer;
  Dtype : char; //used in the case statement to determine if sata or nvme
begin
  //debug
  if debug then DebugOut(true,'getSerials');
  //end debug
  for i:= 1 to dcount do
      begin
        drives[i].serial:='';
        if pos('sr',drives[i].mountpt) > 0 then Dtype:= 'X' //drive is CD so not interested in serial #
        else Dtype:= upcase(drives[i].mountpt[1]); //get first char so can use with case statement
        case Dtype of
             'S' : drives[i].serial:= SataSerial(drives[i].mountpt);
             'N' : drives[i].serial:= NVMEserial(drives[i].mountpt);
        else drives[i].serial:= '';//on case statement
        end;
      end;
end; //end procedure getserials
{-------------------------------------------------------------------------------
FUNCTION - SATASERIAL
Returns the serial number of a sata drive as a string. It uses the utility hdparm, e.g.

pkexec hdparm -I /dev/sda | grep "Serial Number:"

Parameters:
  MountPoint - the mount point for a drive, e.g. sda or sdb

Used by:
  getSerials

V1.0 - 06 June 2020
V1.1 - 05 Sept 2020 - warning removed on failure to get a serial # on a sata device. Seemed to be
                      confusing some people. Would come up everytime someone plugged in a USB stick.
-------------------------------------------------------------------------------}
function SataSerial(MountPoint : string) : string;
const
  bufsize = 64; // Buffer size for reading output
var
  hdProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
begin
  //debug
  if debug then DebugOut(false,'SataSerial','Drive',MountPoint);
  //end debug
  setlength(buffer,bufsize);
  termCmd('pkexec hdparm -I /dev/' + MountPoint + ' | grep "Serial Number:"',hdProcess);
  BytesRead:= hdProcess.output.Read(buffer[1],length(buffer));
  //should get the serial number in one hit
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       SataSerial:= trim(ReplaceStr(buffer,'Serial Number:',''));
     end
  else SataSerial:= '';
  hdProcess.free;
end; //end function sataserial
{-------------------------------------------------------------------------------
FUNCTION - NVMESERIAL
Returns the serial number of an NVME drive as a string. It uses the utility hdparm, e.g.

sudo nvme id-ctrl /dev/nvme0n1 | grep ^sn

Parameters:
  MountPoint - the mount point for a drive, e.g. nvme0n1

Used by:
  getSerials

V1.0 - 06 June 2020
-------------------------------------------------------------------------------}
function NVMESerial(MountPoint : string) : string;
const
  bufsize = 64; // Buffer size for reading output
var
  nvmeProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
begin
  //debug
  if debug then DebugOut(false,'NVMESerial','Drive',MountPoint);
  //end debug
  setlength(buffer,bufsize);
  termCmd('pkexec nvme id-ctrl /dev/' + MountPoint + ' | grep ^sn',nvmeProcess);
  BytesRead:= nvmeProcess.output.Read(buffer[1],length(buffer));
  //should get the serial number in one hit
  if BytesRead > 0 then
     begin
       setlength(buffer,BytesRead);
       NVMESerial:= trim(ReplaceStr(buffer,'sn      :',''));
     end
  else
    begin
      MessageDlg('Warning','Unable to get serial number for ' + MountPoint,mtWarning,[mbOK],0);
      NVMESerial:= '';
    end;
  nvmeProcess.free;
end; //end function NVMEserial

{-------------------------------------------------------------------------------
FUNCTION - ISBOOT
Previous checks on whether the selected drive is the boot drive check to see if sdXY is mounted as /.
This is fine for a standard sata drive including those connected via USB. It fails when the boot drive
is a USB stick containing the foxclone ISO because / is not mounted as, say, sda1. Example output of df
with foxclone booting from a usb stick shows

Filesystem     1K-blocks   Used Available Use% Mounted on
udev             3876000      0   3876000   0% /dev
tmpfs             782704   1172    781532   1% /run
/dev/sdd          619520 619520         0 100% /cdrom
/dev/loop0        532608 532608         0 100% /rofs
/cow             3913504  28644   3884860   1% /
tmpfs            3913504      0   3913504   0% /dev/shm
tmpfs               5120      0      5120   0% /run/lock
tmpfs            3913504      0   3913504   0% /sys/fs/cgroup
tmpfs            3913504      4   3913500   1% /tmp
tmpfs             782700      4    782696   1% /run/user/1000

So, say sdd1, is never found mounting as /.

This function checks for the drive mounting as /cdrom as in:

/dev/sdd          619520 619520         0 100% /cdrom

It also does the previous check - is any partition on sdd mounting as /. For this it relies on other
code populating the global var BootDrive with the mountpoint for the drive. This is currently done in
readdrives in mainform.

Parameters:
  MountPoint - the mount point for a drive, e.g. nvme0n1 or sda

Used by: mainform : backbox1

V1.0 - 09 Sept 2020
-------------------------------------------------------------------------------}
function IsBoot(MountPoint : string) : boolean;
const
  bufsize = 64; // Buffer size for reading output
var
  BootProcess : TProcess;
  Buffer       : string;
  BytesRead    : integer;
begin
  //debug
  if debug then DebugOut(false,'IsUSBboot','Drive',MountPoint);
  //end debug
  if MountPoint = BootDrive then IsBoot:= true else
    begin
      setlength(buffer,bufsize);
      termCmd('df /dev/' + MountPoint + ' | grep /cdrom',BootProcess);
      BytesRead:= BootProcess.output.Read(buffer[1],length(buffer));
      //should get the output number in one hit
      setlength(buffer,BytesRead);
      if pos('/cdrom',buffer) > 0 then IsBoot:= true else IsBoot:= false;
      BootProcess.free;
    end;
end; //end function IsUSBboot

{-------------------------------------------------------------------------------
PROCEDURE PARSEPARTDATA
Introduced in V42, previously the text string for a partition was stored as a string in a TRestData
record. This gives no flexibility in the presentation of the partition information in the UI.

ParsePartData takes the partition text string in a .backup file and parses it into a TRestPartData
record. This can be subsequently read and partition data displayed in a flexible manner, i.e. it can
be changed easily.

See also:
  WritePartInfo - writes partition info into a .backup file
  PartInfo - generates a string from a Tpartparams record for display in check/listboxes
  BuildPartStr - generates a string from a TRestPartData record for display in check/listboxes

Parameters:
  PartStr - a string in the form:
  for gpt
      Partition:sda1 : ntfs : Basic data partition : 542MB : 73% used [Recovery] : Flags-hidden, diag
  for msdos
      Partition:sda1 : ext4 : primary : 14.7GB : 42% used : Flags-boot *Linux Mint 19.3 Tricia*


Used by:  getRestData

V1.0 - 18 Sept 2020
-------------------------------------------------------------------------------}
procedure ParsePartData(PartStr : String; var PartData : TRestPartData);
var
  start, finish : integer;
begin
  //debug
  if debug then DebugOut(true,'ParsePartData','Part string',PartStr);
  //end debug
  //strip out 'Partition:'
  PartStr:= copy(PartStr,pos(':',PartStr) + 1,length(PartStr));
  with PartData do
       begin
         //get the part number, e.g. sda1
         pnumber:= trim(copy(PartStr,1,pos(':',PartStr)-1));
         PartStr:= copy(PartStr,pos(':',PartStr) + 1,length(PartStr)); //get rid of pnumber in partstr
         //if it is an extended partition then that is all that is left
         if pos('extended',PartStr) > 0 then ptype:= 'extended' else
           begin
             //get the filesystem, e.g. ext4
             fs:= trim(copy(PartStr,1,pos(':',PartStr)-1));
             PartStr:= copy(PartStr,pos(':',PartStr) + 1,length(PartStr)); //get rid of fs in partstr
             //get the label - starts with [ and ends with ]
             if pos('[',PartStr) > 0 then //there is a label
                begin
                  start:= pos('[',PartStr);
                  finish:= pos(']',PartStr);
                  plabel:= trim(copy(PartStr,start,finish-start+1));
                  PartStr:= ReplaceStr(PartStr,PartData.plabel,'');
                  end
                  else plabel:= '';
             //get the OS - if present starts with * is the last entry in partstr
             if pos('*',PartStr) > 0 then //there is an OS
                begin
                  OS:= trim(copy(PartStr,pos('*',PartStr),99));//there is an OS
                  PartStr:= ReplaceStr(PartStr,OS,'');
                end
             else OS:= '';
             //different msdos and gpt, next is either the part name (gpt) or type (msdos)
             ptype:= trim(copy(PartStr,1,pos(':',PartStr)-1));
             if not ((ptype = 'primary') or (ptype = 'logical')) then
                begin
                  pname:= ptype;
                  ptype:= '<none>'; //it's an msdos part table, no partition name
                end;
              PartStr:= copy(PartStr,pos(':',PartStr) + 1,length(PartStr)); //get rid of ptype in partstr
              //get the size
             psize:= trim(copy(PartStr,1,pos(':',PartStr)-1));
             PartStr:= copy(PartStr,pos(':',PartStr) + 1,length(PartStr)); //get rid of psize in partstr
             //get percent used - if colon present then there are flags
             if pos(':',PartStr) > 0 then
                begin
                  pused:= trim(copy(PartStr,1,pos(':',PartStr)-1));
                  PartStr:= copy(PartStr,pos(':',PartStr) + 1,length(PartStr)); //get rid of pused in partstr
                  //get flags
                  flags:= trim(PartStr);
                end
             else
               begin
                 pused:= trim(PartStr);
                 flags:= '';
               end;
           end;
       end;
end; //end procedure ParsePartData
{-------------------------------------------------------------------------------
FUNCTION BUILDPARTSTR
Introduced in V42, previously the text string for a partition was stored as a string in a TRestData
record. It allows flexibility in the presentation of partition information.

This takes partition data stored in a TRestPartData record and returns it as a string for display
in partition list/checkboxes. Currently output is generated in the form:

Mountpoint, e.g. sda1
file system, e.g. ext4
[label] if one
size
% used
flags if any
type, i.e. primary, extended, logical (legacy only) or
(partition name)
*OS* if present (and adding a * to the start of the string returned)

Parameters:
  PartData - a TRestPartData record previously populated by ParsePartData

Used by:

V1.0 - 19 Sept 2020
V1.1 - 15 Jan 2021 : error had not included the partdata output within debug
-------------------------------------------------------------------------------}
function BuildPartStr(PartData : TRestPartData) : string;
begin
  //debug
  if debug then DebugOut(false,'BuildPartStr','Part string for',PartData.pnumber);
  if debug then with partdata do
       begin
         writeln('** part = ',pnumber);
         writeln('** fs = ',fs);
         writeln('** size = ',psize);
         writeln('** pused = ',pused);
         writeln('** pname = ',pname);
         writeln('** label = ',plabel);
         writeln('** flags = ',flags);
         writeln('** ptype = ',ptype);
         writeln('** OS = ',OS);
         writeln('** blocks = ',IntToStr(blocks));
       end;
  //end debug
  BuildPartStr:= '';
  with PartData do
       begin
         if length(OS) > 0 then BuildPartStr:= '*';
         BuildPartStr+= pnumber + ' : ';
         if ptype = 'extended' then BuildPartStr+= ptype else
           begin
             BuildPartStr+= fs + ' : ';
             BuildPartStr+= plabel + ' : ';
             BuildPartStr+= psize + ' : ';
             BuildPartStr+= pused + ' : ';
             BuildPartStr+= flags + ' : ';
             //can either have a pname = gpt part table or part type = msdos
             if (ptype = 'primary') or (ptype = 'logical') then BuildPartStr+= ptype + ' ' else
                if pname <> '<none>' then BuildPartStr+= '(' + pname + ') ';
             BuildPartStr+= OS;
           end;
         //get rid of the UUID in a swap line
         if pos('swap',BuildPartStr) > 0 then BuildPartStr:= pnumber + ' : swap';
       end;
  //tidy up
  BuildPartStr:= replacestr(BuildPartStr,': :','::');
 end; //end procedure ParsePartData
 {-------------------------------------------------------------------------------
FUNCTION WRITEPARTINFO
Generates a string of partition info for display in partition checkboxes. This is the original partinfo
function. It was modified to change the order of information in a partition checkbox based on user
feedback. But it was also used to write the partition information to the .backup file when doing a
backup. This CANNOT be changed without losing backwards compatability with ALL previous backups under
previous versions.

This is ONLY used by BackBackup, other procedures that display partition info use the modified
partinfo function.

Parameters:
  partition - a Tpartparams record containing the partition info to be parsed
  parttype - a string defining the partition table type as either MSDOS or GPT.
             this is required because the information returned by the function
             differs between MSDOS and GPT.

Used by:
  Unit backup - BackBackup

V1.0 - 19 Sept 2020
-------------------------------------------------------------------------------}
function WritePartInfo(var partition : Tpartparams; parttype : string) : string;
const
  ADJUST : real = 1.024; //parted is using 1024 for blocks, df 1000
  MULTIPLY = 1000;
var
  tempnum : real;
begin
  //debug
  if debug then DebugOut(false,'WritePartInfo','part record for',partition.pnumber,'parttype',parttype);
  //end debug
  with partition do
       begin
         WritePartInfo:= pnumber + ' : ';
         WritePartInfo+= fs + ' : ';
         if parttype = 'GPT' then WritePartInfo+= pname + ' : '
            else WritePartInfo+= ptype + ' : ';
         if ptype <> 'extended' then
            begin
              if (fs = '<unknown>') or (fs = 'swap') then tempnum:= psize/MULTIPLY
                 else tempnum:= psize/MULTIPLY * ADJUST;
              if tempnum > 100000 then WritePartInfo+= FloatToStrF(tempnum/MULTIPLY,ffFixed,4,0) + 'GB'
                 else if tempnum > 1000 then WritePartInfo+= FloatToStrF(tempnum/MULTIPLY,ffFixed,4,1) + 'GB'
                         else WritePartInfo+= FloatToStrF(tempnum,ffFixed,4,0) + 'MB';
              WritePartInfo+= ' : ' + PercentUsed(pused,psize);
              if plabel <> '<none>' then WritePartInfo+= ' [' + plabel + ']';
              if flags <> '<none>' then WritePartInfo+= ' : Flags-' + flags;
            end;
         if OS <> '' then WritePartInfo:= '*' + WritePartInfo + ' *' + OS + '*';
       end;
end; //end function WritePartInfo
{-------------------------------------------------------------------------------
PROCEDURE SWAPON
Introduced in V42, either turns swap on or swap off where the action is determined by OnOff. Previously
swap was turned off in either the iso or the deb. This procedure brings it within foxclone. This is
required because on boot systemd will use any swap partition it finds. If this is on a partition that
is either being restored or cloned to, sfdisk will be unable to complete correctly - the kernel will not
be able to re-read the partition table because it is in use. The net result is that the restore/clone
will complete but it won't work. This is needed more for the deb than the iso.

Note - needs a polkit for swapon and swapoff.

Parameters:
  OnOff = true then turn swap on, if false turn swap off.

Used by: foxcloneXX

V1.0 - 03 Oct 2020
-------------------------------------------------------------------------------}
procedure SwapOn(OnOff : boolean);
const
  bufsize = 64; // Buffer size for reading output
var
  swapProcess           : TProcess;
  buffer,setting,action : string;
  BytesRead             : integer;
begin
  //debug
  if debug then
     begin
       if OnOff then setting:= 'ON' else setting:= 'OFF';
       DebugOut(true,'Swap','Setting',setting);
     end;
  //end debug
  setlength(buffer,bufsize);
  if OnOff then action:= 'pkexec swapon -a' else action:= 'pkexec swapoff -a';
  termCmd(action,swapProcess);
  //assume output in one hit
  BytesRead:= swapProcess.Output.Read(buffer[1],length(buffer));
  swapProcess.free;
end; //end procedure SwapOn
{-------------------------------------------------------------------------------
FUNCTION PARTCLONEV
Builds the command line for running a partclone verify based the backup filename
(to extract the name out of the *.backup name). The command line is then built
based on settings in VSettings.

Note the used of the -F option to force the restore, partclone was throwing up
an error "Target partition size(0 MB) is smaller than source(20972 MB)" on an
ext4 extended partition (was quite happy with primary partitions in MSDOS &
GPT part tables). Checked with sfdisk and the restored partition table was
NOT showing a zero size partition. Same is also true on the clone procedures.

Test - ferens OS with /home in a logical part on mbr drive. Restored okay with
-F option, but then failed to mount when foxclone run again. Reboot into ferens,
everything okay, its /home is fine. Reboot into LM19 run foxclone, no problems
reading sda and the logical part giving problems????? Not something that would be
encountered normally - after a restore, the next action would be a reboot.

Returns a command line that can be executed by bash.

Parameters:
  RestFname  - is a string and is of the form *.backup and is the FULL pathname
  PartNum - a string of the form sda1 and used to build up the filename for the image file.

               dont like the use of destmountpt it is hiding what is going on and not obvious werhe set

Used by: Unit verify - VerifyParts

V1.0 - 14 Jan 2021
-------------------------------------------------------------------------------}
function PartCloneV(RestFname : string; PartNum : string) : string;
var
  Fname : string;
  //added in -F = force cos of problems with extended part.
  POptions : string = ' -s - -L ';
  Logfile : string;
begin
  //debug
  if debug then DebugOut(false,'PartCloneV','RestFname',RestFname,'Partition',PartNum);
  //end debug
  //get backup filename, add in the partition number and add .img
  Fname:= ReplaceStr(RestFname,'backup',PartNum);
  Fname:= Fname + '.img';
  //escape any spaces in the filename
  if pos(' ',Fname) > 0 then Fname:= StringReplace(Fname,' ','\ ',[rfReplaceAll]);
  //replace the mount point, e.g. sda1 in the backup with the mount point on the destination,
  //e.g. the drive may have moved from sda when the backup was done to sdx.
  //PartNum:= '/dev/' + PartNum;
  //just in case the drive has moved its mount point between the backup and the restore
  //PartNum:= ReplaceStr(PartNum,RSettings.Mountpt,RSettings.DestMountpt);
  PartCloneV:= 'partclone.chkimg';
  //build command line
  Logfile:= Fname + '.chk';
  with VSettings do
       if Compress and Split then
          // 20190603.sda3.img.gz.01
          // cat 20190603.sda3.img.gz.* | unpigz -c | partclone.chkimg -s - -L logfile
          PartCloneV:= 'cat ' + Fname + '.gz.* | unpigz -c | ' + PartCloneV + POptions + Logfile
          else if Compress and not Split then
               // 20190604.sda3.img.gz
               // cat 20190603.sda3.img.gz | unpigz -c | partclone.chkimg -s - -L logfile
               PartCloneV:= 'cat ' + Fname + '.gz | unpigz -c | ' + PartCloneV + POptions + Logfile
               else if not Compress and Split then
                    // 20190604.sda3.img.00
                    // note '-s -' to read input from stream not file
                    // cat 20190603.sda3.img.* | unpigz -c | partclone.chkimg -s - -L logfile
                    PartCloneV:= 'cat ' + Fname + '.* | ' + PartCloneV + POptions + Logfile
                    else
                      // 20190604.sda3.img
                      // partclone.chkimg 20190604.sda3.img -L Logfile
                      PartCloneV:= PartCloneV + ' -s ' + Fname + ' -L ' + Logfile;
end; //end function partcloneV

end.

